from __future__ import annotations

import shutil
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

from aicd.agent.tool_args_coerce import coerce_tool_json_array
from aicd.runtime_task_context import current_task_id
from aicd.tasks.main_inbox import (
    MAIN_INBOX_PREFIX,
    is_main_inbox_task_id,
    main_inbox_id,
)
from aicd.tasks.safe_ids import is_safe_task_id, validate_topic_slug

if TYPE_CHECKING:
    from aicd.agent.tool_router import ToolRouter

_ALLOWED_MSG_KINDS = frozenset(
    {
        "handoff",
        "status",
        "question",
        "artifact_index",
        "alert",  # 子任务向上级报告异常/风险（与 status 区分：更需尽快处理）
        "directive",  # 上级/主会话对用户新意图的转发，子 Agent 收件箱驱动继续执行
        "lifecycle",  # 运行时写入主收件箱：顶层子任务完成/失败/取消（也可由工具显式发送）
    }
)


def _parent_id_from_context() -> str | None:
    return current_task_id.get()


async def _resolve_thread_root_for_task_chain(
    router: ToolRouter, start_task_id: str
) -> str | None:
    """沿 parent 链查找 payload.thread_root_id（嵌套子任务对齐主会话）。"""
    seen: set[str] = set()
    tid = start_task_id.strip()
    while tid and tid not in seen:
        seen.add(tid)
        row = await router._tasks.get_task_row(tid)
        if row is None:
            break
        tr = row.payload.get("thread_root_id")
        if isinstance(tr, str) and tr.strip():
            return tr.strip()
        pid = row.parent_id
        tid = pid.strip() if isinstance(pid, str) and pid.strip() else ""
    return None


async def _task_shell(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    parent_id = _parent_id_from_context()
    thread_root_id: str | None = None
    if parent_id is None:
        sid = router._chat_session_id
        if isinstance(sid, str) and sid.strip():
            thread_root_id = sid.strip()
    try:
        tid = await router._tasks.create_shell_task(
            args["command"],
            parent_id=parent_id,
            cwd=args.get("cwd"),
            thread_root_id=thread_root_id,
        )
    except ValueError as e:
        return {"ok": False, "error": str(e)}
    return {
        "ok": True,
        "task_id": tid,
        "note": "Use task_result for payload.shell_process (root_pid of the shell wrapper). "
        "On Windows, task_cancel runs taskkill /T on that pid so GUI children exit too.",
    }


async def _task_dummy(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    parent_id = _parent_id_from_context()
    thread_root_id: str | None = None
    if parent_id is None:
        sid = router._chat_session_id
        if isinstance(sid, str) and sid.strip():
            thread_root_id = sid.strip()
    tid = await router._tasks.create_dummy_long_task(
        steps=int(args.get("steps", 5)),
        parent_id=parent_id,
        thread_root_id=thread_root_id,
    )
    return {"ok": True, "task_id": tid}


async def _task_ai(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    parent_id = _parent_id_from_context()
    thread_root_id: str | None = None
    if parent_id is None:
        sid = router._chat_session_id
        if isinstance(sid, str) and sid.strip():
            thread_root_id = sid.strip()

    ro = args.get("read_only_paths")
    ro_list: list[str] | None = None
    if isinstance(ro, list):
        ro_list = [str(x) for x in ro if isinstance(x, str) and x.strip()]
        if not ro_list:
            ro_list = None

    topic_slug = args.get("topic_slug")
    ts_raw = topic_slug if isinstance(topic_slug, str) else None
    try:
        ts = validate_topic_slug(ts_raw)
    except ValueError as e:
        return {"ok": False, "error": str(e)}

    ph = args.get("parent_handoff_expectation")
    ph_exp = ph if isinstance(ph, str) else None

    disp = args.get("display_title")
    disp_s = str(disp).strip()[:200] if isinstance(disp, str) and disp.strip() else None
    purp = args.get("purpose")
    purp_s = str(purp).strip()[:4000] if isinstance(purp, str) and purp.strip() else None
    est = args.get("estimated_steps")
    est_i: int | None = None
    if est is not None:
        try:
            est_i = int(est)
        except (TypeError, ValueError):
            est_i = None

    try:
        tid = await router._tasks.create_ai_task(
            args["instruction"],
            parent_id=parent_id,
            topic_slug=ts,
            read_only_paths=ro_list,
            parent_handoff_expectation=ph_exp,
            thread_root_id=thread_root_id,
            display_title=disp_s,
            purpose=purp_s,
            estimated_steps=est_i,
        )
    except ValueError as e:
        return {"ok": False, "error": str(e)}
    return {"ok": True, "task_id": tid}


def _task_timeline_payload(p: dict[str, Any]) -> dict[str, Any]:
    keys = (
        "started_at",
        "finished_at",
        "duration_sec",
        "progress",
        "sub_agent_trace",
        "display_title",
        "purpose",
        "estimated_steps",
        "shell_process",
    )
    return {k: p[k] for k in keys if k in p and p[k] is not None}


async def _task_list(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    rows = await router._tasks.list_tasks(limit=int(args.get("limit", 30)))
    tm = router._tasks
    return {
        "ok": True,
        "tasks": [
            {
                "id": r.id,
                "type": r.type,
                "status": r.status,
                "running_now": tm.is_task_running(r.id),
                "parent_id": r.parent_id,
                "created_at": r.created_at,
                "updated_at": r.updated_at,
                "summary": r.result_summary,
                "error": r.error,
                "timeline": _task_timeline_payload(r.payload),
            }
            for r in rows
        ],
    }


async def _task_cancel(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    return await router._tasks.cancel_task(args["task_id"])


async def _task_result(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    r = await router._tasks.get_task_row(args["task_id"])
    if not r:
        return {"ok": False, "error": "not found"}
    tm = router._tasks
    return {
        "ok": True,
        "task": {
            "id": r.id,
            "type": r.type,
            "status": r.status,
            "running_now": tm.is_task_running(r.id),
            "parent_id": r.parent_id,
            "created_at": r.created_at,
            "updated_at": r.updated_at,
            "payload": r.payload,
            "timeline": _task_timeline_payload(r.payload),
            "result_summary": r.result_summary,
            "error": r.error,
        },
    }


def _task_matches_thread(row: Any, thread_root_id: str) -> bool:
    if row.id == thread_root_id:
        return True
    p = row.payload or {}
    tr = p.get("thread_root_id")
    return isinstance(tr, str) and tr.strip() == thread_root_id


async def _task_progress(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    tid_arg = args.get("task_id")
    if isinstance(tid_arg, str) and tid_arg.strip():
        tid = tid_arg.strip()
    else:
        ctx = current_task_id.get()
        tid = ctx.strip() if isinstance(ctx, str) and ctx.strip() else ""
    if not tid:
        return {
            "ok": False,
            "error": "task_id required when not running inside a task_ai worker",
        }
    if not is_safe_task_id(tid):
        return {"ok": False, "error": "invalid task_id"}
    row = await router._tasks.get_task_row(tid)
    if not row:
        return {"ok": False, "error": "unknown task_id"}
    try:
        step_index = int(args["step_index"])
    except (KeyError, TypeError, ValueError):
        return {"ok": False, "error": "step_index (integer) is required"}
    step_total = args.get("step_total")
    st: int | None = None
    if step_total is not None:
        try:
            st = int(step_total)
        except (TypeError, ValueError):
            st = None
    cf = args.get("current_focus")
    cf_s = str(cf).strip()[:2000] if isinstance(cf, str) else ""
    note = args.get("note")
    note_s = str(note).strip()[:2000] if isinstance(note, str) else ""
    prog = {
        "step_index": step_index,
        "updated_at": datetime.now(timezone.utc).isoformat(),
    }
    if st is not None and st > 0:
        prog["step_total"] = st
    if cf_s:
        prog["current_focus"] = cf_s
    if note_s:
        prog["note"] = note_s
    await router._tasks.patch_task_payload(tid, {"progress": prog})
    await router._tasks._emit(
        tid,
        "progress",
        phase="explicit",
        step_index=step_index,
        step_total=st,
        current_focus=cf_s[:240],
        message=(
            f"进度 {step_index}"
            + (f"/{st}" if st else "")
            + (f" — {cf_s[:120]}" if cf_s else "")
        ),
    )
    return {"ok": True, "task_id": tid, "progress": prog}


async def _task_thread_stats(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    tr = args.get("thread_root_id")
    tr_s = tr.strip() if isinstance(tr, str) and tr.strip() else ""
    if not tr_s:
        sid = router._chat_session_id
        tr_s = sid.strip() if isinstance(sid, str) and sid.strip() else ""
    if not tr_s:
        return {
            "ok": False,
            "error": "thread_root_id required (or main chat session unavailable)",
        }
    rows = await router._tasks.list_tasks(limit=int(args.get("limit", 250)))
    matched = [r for r in rows if _task_matches_thread(r, tr_s)]
    running = sum(1 for r in matched if r.status == "running")
    completed = sum(1 for r in matched if r.status == "completed")
    failed = sum(1 for r in matched if r.status == "failed")
    cancelled = sum(1 for r in matched if r.status == "cancelled")
    sum_dur = 0.0
    for r in matched:
        ds = r.payload.get("duration_sec")
        if isinstance(ds, (int, float)):
            sum_dur += float(ds)
    earliest: str | None = None
    latest_finish: str | None = None
    for r in matched:
        sa = r.payload.get("started_at")
        if isinstance(sa, str) and sa.strip():
            if earliest is None or sa < earliest:
                earliest = sa
        fa = r.payload.get("finished_at")
        if isinstance(fa, str) and fa.strip():
            if latest_finish is None or fa > latest_finish:
                latest_finish = fa
    wall_sec: float | None = None
    if earliest and latest_finish:
        try:
            t0 = datetime.fromisoformat(earliest.replace("Z", "+00:00"))
            t1 = datetime.fromisoformat(latest_finish.replace("Z", "+00:00"))
            if t0.tzinfo is None:
                t0 = t0.replace(tzinfo=timezone.utc)
            if t1.tzinfo is None:
                t1 = t1.replace(tzinfo=timezone.utc)
            wall_sec = round((t1 - t0).total_seconds(), 3)
        except ValueError:
            wall_sec = None
    return {
        "ok": True,
        "thread_root_id": tr_s,
        "task_count": len(matched),
        "by_status": {
            "running": running,
            "completed": completed,
            "failed": failed,
            "cancelled": cancelled,
        },
        "sum_duration_sec": round(sum_dur, 3),
        "wall_span_sec": wall_sec,
        "earliest_started_at": earliest,
        "latest_finished_at": latest_finish,
    }


def _serialize_task_message(m: Any) -> dict[str, Any]:
    return {
        "id": m.id,
        "thread_root_id": m.thread_root_id,
        "to_task_id": m.to_task_id,
        "from_task_id": m.from_task_id,
        "kind": m.kind,
        "body": m.body,
        "created_at": m.created_at,
        "read_at": m.read_at,
    }


async def _task_message_send(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    to_raw = str(args["to_task_id"]).strip()
    kind = str(args["kind"]).strip()
    if kind not in _ALLOWED_MSG_KINDS:
        return {
            "ok": False,
            "error": f"kind must be one of: {sorted(_ALLOWED_MSG_KINDS)}",
        }
    body = args.get("body")
    if body is None:
        body_d: dict[str, Any] = {}
    elif isinstance(body, dict):
        body_d = body
    else:
        return {"ok": False, "error": "body must be a JSON object"}

    from_tid = args.get("from_task_id")
    if isinstance(from_tid, str) and from_tid.strip():
        fid = from_tid.strip()
    else:
        ctx = current_task_id.get()
        fid = ctx.strip() if isinstance(ctx, str) and ctx.strip() else ""
    if not fid:
        sid = router._chat_session_id
        if isinstance(sid, str) and sid.strip():
            fid = f"main-{sid.strip()}"
        else:
            return {
                "ok": False,
                "error": "from_task_id required when not in task_ai and no chat session",
            }

    resolved_to = to_raw
    if to_raw.lower() == "main":
        ctx2 = current_task_id.get()
        if ctx2 and str(ctx2).strip():
            trt = await _resolve_thread_root_for_task_chain(
                router, str(ctx2).strip()
            )
            if not trt:
                return {
                    "ok": False,
                    "error": "无法解析主会话：当前任务链上缺少 payload.thread_root_id",
                }
            resolved_to = main_inbox_id(trt)
        else:
            msid = router._chat_session_id
            if not (isinstance(msid, str) and msid.strip()):
                return {
                    "ok": False,
                    "error": "主对话发送 to_task_id=main 需要有效 chat session",
                }
            resolved_to = main_inbox_id(msid.strip())
    elif is_main_inbox_task_id(to_raw):
        resolved_to = to_raw

    if not is_safe_task_id(resolved_to) or not is_safe_task_id(fid):
        return {"ok": False, "error": "invalid task id format"}

    trow = await router._tasks.get_task_row(resolved_to)
    if not is_main_inbox_task_id(resolved_to):
        if not trow:
            return {"ok": False, "error": "unknown to_task_id"}

    tr = args.get("thread_root_id")
    thread_root: str | None = None
    if isinstance(tr, str) and tr.strip():
        thread_root = tr.strip()
    if thread_root is None:
        srow = await router._tasks.get_task_row(fid)
        if srow:
            tr2 = srow.payload.get("thread_root_id")
            if isinstance(tr2, str) and tr2.strip():
                thread_root = tr2.strip()
    if thread_root is None and isinstance(fid, str) and fid.startswith("main-"):
        rest = fid[5:].strip()
        if rest:
            thread_root = rest
    if thread_root is None and trow is not None:
        tr3 = trow.payload.get("thread_root_id")
        if isinstance(tr3, str) and tr3.strip():
            thread_root = tr3.strip()
    if thread_root is None and is_main_inbox_task_id(resolved_to):
        thread_root = resolved_to[len(MAIN_INBOX_PREFIX) :]

    mid = await router._tasks.post_task_message(
        to_task_id=resolved_to,
        from_task_id=fid,
        kind=kind,
        body=body_d,
        thread_root_id=thread_root,
    )
    return {
        "ok": True,
        "message_id": mid,
        "resolved_to_task_id": resolved_to,
    }


async def _task_message_inbox(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    to_arg = args.get("to_task_id")
    if isinstance(to_arg, str) and to_arg.strip():
        to_tid = to_arg.strip()
    else:
        ctx = current_task_id.get()
        to_tid = ctx.strip() if isinstance(ctx, str) and ctx.strip() else ""
    if not to_tid:
        return {
            "ok": False,
            "error": "to_task_id required when not running inside a task_ai worker",
        }
    if not is_safe_task_id(to_tid):
        return {"ok": False, "error": "invalid to_task_id"}
    since = int(args.get("since_id", 0))
    limit = min(int(args.get("limit", 50)), 200)
    rows = await router._tasks.store.list_task_messages(
        router._tasks.db_conn,
        to_task_id=to_tid,
        since_id=since,
        limit=limit,
    )
    return {
        "ok": True,
        "to_task_id": to_tid,
        "messages": [_serialize_task_message(m) for m in rows],
    }


async def _task_message_mark_read(
    router: ToolRouter, args: dict[str, Any]
) -> dict[str, Any]:
    raw = args.get("message_ids")
    co = coerce_tool_json_array(raw, "message_ids", allow_empty=False)
    if isinstance(co, dict):
        return co
    raw = co
    ids: list[int] = []
    for x in raw:
        try:
            ids.append(int(x))
        except (TypeError, ValueError):
            return {"ok": False, "error": "message_ids must be integers"}
    await router._tasks.store.mark_task_messages_read(router._tasks.db_conn, ids)
    return {"ok": True, "marked": len(ids)}


async def _task_main_inbox(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    """主对话：拉取 main-inbox-<session> 中的未读/增量 task_message。"""
    if current_task_id.get() is not None:
        return {
            "ok": False,
            "error": "task_main_inbox 仅可在主对话使用；子 Agent 请用 task_message_send(..., to_task_id=\"main\", ...)",
        }
    sid = router._chat_session_id
    if not (isinstance(sid, str) and sid.strip()):
        return {"ok": False, "error": "无 chat session，无法解析主收件箱"}
    to_id = main_inbox_id(sid.strip())
    since = int(args.get("since_message_id", 0) or 0)
    limit = min(max(int(args.get("limit", 50) or 50), 1), 200)
    rows = await router._tasks.store.list_task_messages(
        router._tasks.db_conn,
        to_task_id=to_id,
        since_id=since,
        limit=limit,
    )
    mark = args.get("mark_read")
    do_mark = True if mark is None else bool(mark)
    if do_mark and rows:
        await router._tasks.store.mark_task_messages_read(
            router._tasks.db_conn, [m.id for m in rows]
        )
    return {
        "ok": True,
        "main_inbox_id": to_id,
        "since_message_id": since,
        "messages": [_serialize_task_message(m) for m in rows],
        "note": "子任务结束时会自动写入 kind=lifecycle；嵌套子任务上行请 to_task_id=\"main\"。"
        " 定时检查周期用 **agent.idleHeartbeatSec**（TUI `/agent heartbeat`）或本工具主动拉取。",
    }


async def _task_session_children_digest(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    """主对话：列出本会话发起的顶层后台任务（parent_id 为空且 thread_root_id 匹配）。"""
    if current_task_id.get() is not None:
        return {
            "ok": False,
            "error": "task_session_children_digest 仅可在主对话使用",
        }
    sid = router._chat_session_id
    if not (isinstance(sid, str) and sid.strip()):
        return {"ok": False, "error": "无 chat session"}
    since_u = args.get("since_updated_at")
    since_s = str(since_u).strip() if since_u is not None and str(since_u).strip() else None
    limit = min(max(int(args.get("limit", 40) or 40), 1), 200)
    rows = await router._tasks.store.list_top_level_tasks_for_session_thread(
        router._tasks.db_conn,
        sid.strip(),
        limit=limit,
        since_updated_at=since_s,
    )
    return {
        "ok": True,
        "session_thread_id": sid.strip(),
        "since_updated_at": since_s,
        "tasks": [
            {
                "id": r.id,
                "type": r.type,
                "status": r.status,
                "running_now": router._tasks.is_task_running(r.id),
                "updated_at": r.updated_at,
                "summary": r.result_summary,
                "error": r.error,
                "display_title": (r.payload or {}).get("display_title"),
            }
            for r in rows
        ],
        "note": "仅包含带 payload.thread_root_id 的顶层任务（主对话发起的 task_ai/task_shell 等）。"
        " 若 since_updated_at 为 ISO 时间字符串，只返回 updated_at 更晚的行。",
    }


async def _task_history_clear(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    if current_task_id.get() is not None:
        return {
            "ok": False,
            "error": "task_history_clear 仅可在主对话中使用，不能在 task_ai 子 Agent 内调用",
        }
    if args.get("confirm") is not True:
        return {
            "ok": False,
            "error": "须设置 confirm=true 才会清空：SQLite 中全部 tasks/task_messages、"
            "AICD_HOME/tasks/<id> 工作区目录，以及当前 AI tmp/tasks/* 草稿目录。",
        }
    return await router._tasks.clear_task_history()


async def _task_scratch_cleanup(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    tid_arg = args.get("task_id")
    if isinstance(tid_arg, str) and tid_arg.strip():
        tid = tid_arg.strip()
    else:
        ctx = current_task_id.get()
        tid = ctx.strip() if isinstance(ctx, str) and ctx.strip() else ""
    if not tid:
        return {
            "ok": False,
            "error": "task_id required when not running inside a task_ai worker",
        }
    if not is_safe_task_id(tid):
        return {"ok": False, "error": "invalid task_id"}

    _out, ai_tmp = router.get_ai_paths()
    base = (ai_tmp / "tasks").resolve()
    root = (base / tid).resolve()
    try:
        root.relative_to(base)
    except ValueError:
        return {"ok": False, "error": "scratch path escaped allowed directory"}
    if not root.is_dir():
        root.mkdir(parents=True, exist_ok=True)
        return {"ok": True, "cleaned": str(root), "note": "was missing; created empty"}
    shutil.rmtree(root)
    root.mkdir(parents=True, exist_ok=True)
    return {"ok": True, "cleaned": str(root)}


HANDLERS: dict[str, Any] = {
    "task_shell": _task_shell,
    "task_dummy": _task_dummy,
    "task_ai": _task_ai,
    "task_list": _task_list,
    "task_cancel": _task_cancel,
    "task_result": _task_result,
    "task_progress": _task_progress,
    "task_thread_stats": _task_thread_stats,
    "task_history_clear": _task_history_clear,
    "task_message_send": _task_message_send,
    "task_message_inbox": _task_message_inbox,
    "task_message_mark_read": _task_message_mark_read,
    "task_main_inbox": _task_main_inbox,
    "task_session_children_digest": _task_session_children_digest,
    "task_scratch_cleanup": _task_scratch_cleanup,
}



from __future__ import annotations

import asyncio
import os
import re
import shutil
import sys
from pathlib import Path
from typing import Any, Callable, Awaitable

from aicd.cli_encoding import subprocess_env_utf8
from aicd.tools.path_policy import PathPolicy


def resolve_windows_comspec() -> str:
    """``%ComSpec%``，不可用时回退到 PATH 中的 ``cmd.exe``。"""
    for key in ("ComSpec", "COMSPEC"):
        c = os.environ.get(key)
        if c:
            try:
                if Path(c).expanduser().is_file():
                    return str(Path(c).resolve())
            except OSError:
                pass
            return c
    w = shutil.which("cmd.exe")
    return w if w else "cmd.exe"


def windows_cmd_argv(command_line: str) -> list[str]:
    """
    通过 CMD 执行一行命令：``/d`` 禁用 AutoRun；先 ``chcp 65001`` 再执行，与 UTF-8 环境一致。
    ``command_line`` 将传给 ``cmd /c`` 的剩余部分（勿再外包一层引号除非语法需要）。
    """
    comspec = resolve_windows_comspec()
    inner = f"chcp 65001>nul & {command_line}"
    return [comspec, "/d", "/c", inner]


def windows_powershell_argv(command_line: str, *, flavor: str = "powershell") -> list[str]:
    """``flavor``: ``powershell``（Windows PowerShell 5）或 ``pwsh``（PowerShell 7+）。"""
    name = "pwsh.exe" if flavor.lower().strip() == "pwsh" else "powershell.exe"
    exe = shutil.which(name) or shutil.which(name.replace(".exe", "")) or name
    return [
        exe,
        "-NoProfile",
        "-NonInteractive",
        "-ExecutionPolicy",
        "Bypass",
        "-Command",
        command_line,
    ]


def shell_argv_from_user_input(
    *,
    command: str | None = None,
    argv: list[str] | None = None,
    shell: str | None = None,
) -> list[str]:
    """
    - ``argv``：直接 ``execve`` 风格参数列表（不经 CMD/sh 解释），适用于含空格路径、``git``/``python`` 等。
    - ``command``：在 Windows 上经 ``shell``（默认 CMD）执行的一行命令；在 Unix 上为 ``sh -c``。
    """
    if argv is not None:
        if len(argv) == 0:
            raise ValueError("argv must be a non-empty array of strings")
        return [str(x) for x in argv]
    if command is None or not str(command).strip():
        raise ValueError("non-empty command or argv is required")
    line = str(command).strip()
    if sys.platform != "win32":
        return ["/bin/sh", "-c", line]
    sh = (shell or "cmd").strip().lower()
    if sh in ("powershell", "ps"):
        return windows_powershell_argv(line, flavor="powershell")
    if sh == "pwsh":
        return windows_powershell_argv(line, flavor="pwsh")
    return windows_cmd_argv(line)


def default_shell_argv(command: str) -> list[str]:
    """兼容旧调用：等价于 ``shell_argv_from_user_input(command=command)``。"""
    return shell_argv_from_user_input(command=command)


def warn_shell_python_c_unbalanced_quotes(command: str) -> str | None:
    """
    检测一行 shell 里 ``python -c "..."`` 的双引号是否成对（粗略启发式）。
    """
    s = str(command).strip()
    if not re.search(r"(?i)python(?:\.exe)?\b", s):
        return None
    if not re.search(r"(?i)(^|[\s&(])-c(\s+|$)", s):
        return None
    m = re.search(r"(?i)-c\s+(.*)$", s, re.DOTALL)
    if not m:
        return None
    tail = m.group(1).strip()
    if not tail:
        return "python -c 缺少代码参数"
    if tail.startswith("'"):
        return None
    if tail.count('"') % 2 == 1:
        return (
            "检测到 python -c 后的双引号未成对（CMD 下常与 && 混用导致截断，出现 SyntaxError）。"
            "请改用 **script_run**（kind=py）、或 **fs_write_file** 写入 .py 后用 **run_command** 的 "
            "**argv**：[\"python\", \"脚本路径\"]。避免在一条 CMD 链里塞长段 python -c。"
        )
    return None


def warn_shell_windows_python_c_with_and_chain(command: str) -> str | None:
    """
    Windows CMD 下 ``cd ... && python -c "..."`` 即使源码里引号成对，也常把 ``-c`` 实参截断，
    表现为 ``SyntaxError: unterminated string literal``（例如只收到 ``"import``）。
    """
    if sys.platform != "win32":
        return None
    s = str(command).strip()
    if "&&" not in s:
        return None
    if not re.search(r"(?i)python(?:\.exe)?\b", s):
        return None
    if not re.search(r"(?i)-c\s+", s):
        return None
    return (
        "在 Windows CMD 中，请勿把 ``cd ... && python -c \"...\"`` 写在同一行（``&&`` 与双引号组合易截断参数，触发 SyntaxError）。"
        "请改用 **script_run**（kind=py）、分两次 **run_command**（先 cd /d，再 python），"
        "或 **fs_write_file** 写入 .py 后用 **argv**：[\"python\", \"脚本路径\"]。"
    )


_FINDSTR_COMMON_NOEXT_FILES = frozenset(
    x.lower()
    for x in (
        "LICENSE",
        "README",
        "COPYING",
        "AUTHORS",
        "CHANGELOG",
        "Makefile",
        "Dockerfile",
        "Procfile",
        "CONTRIBUTING",
        "GITATTRIBUTES",
        "GITIGNORE",
    )
)


def warn_shell_windows_findstr_maybe_dir_operand(command: str) -> str | None:
    """
    Windows ``findstr`` 把「搜索串」后面的实参当作**要打开的文件**；若写成子目录名而无通配符，
    会得到 ``FINDSTR: Cannot open <name>``（并非文件被锁或未释放）。
    """
    if sys.platform != "win32":
        return None
    s = str(command).strip()
    if not re.search(r'(?i)\bfindstr\b', s):
        return None
    if "*" in s:
        return None
    tail = re.split(r"[&|]+", s)[-1].strip()
    toks = tail.split()
    if len(toks) < 3:
        return None
    last = toks[-1].strip('"')
    if not last or last.startswith("/") or last.startswith("-"):
        return None
    if "." in last or last.endswith("\\") or last.endswith("/"):
        return None
    if last.lower() in _FINDSTR_COMMON_NOEXT_FILES:
        return None
    return (
        "Windows **findstr** 将把最后的参数当作**单个文件**；若为**文件夹**会报 *Cannot open*（不是资源锁）。"
        "请改为 `cd /d <dir> && findstr /s /i <pattern> *.*`，或 `findstr /s /i <pattern> <dir>\\\\*.*`，"
        "或改用 **rg** / **fs_search_regex**。"
    )


def warn_shell_risky_python_c(command: str) -> str | None:
    """合并：未成对引号，或 ``&&`` + ``python -c``（仅 Windows）。"""
    return warn_shell_python_c_unbalanced_quotes(
        command
    ) or warn_shell_windows_python_c_with_and_chain(command)


def warn_shell_preflight(command: str) -> str | None:
    """``task_shell`` / ``run_command`` 提交前：Python 引号风险 + Windows findstr 误用提示。"""
    return warn_shell_risky_python_c(command) or warn_shell_windows_findstr_maybe_dir_operand(
        command
    )


async def kill_windows_process_tree(
    root_pid: int, *, timeout_sec: float = 15.0
) -> dict[str, Any]:
    """
    结束 ``root_pid`` 及其**所有后代**（``taskkill /T /F``）。

    用于 Windows 上通过 CMD/PowerShell 拉起的 GUI 等子进程：仅 ``terminate`` 外层 ``cmd.exe``
    时，子进程常会残留；整树结束后可一并清理。
    """
    pid = int(root_pid)
    try:
        tk = await asyncio.create_subprocess_exec(
            "taskkill",
            "/PID",
            str(pid),
            "/T",
            "/F",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
    except FileNotFoundError:
        return {"ok": False, "error": "taskkill not found on PATH"}
    try:
        out_b, err_b = await asyncio.wait_for(
            tk.communicate(), timeout=timeout_sec
        )
    except TimeoutError:
        tk.kill()
        return {"ok": False, "error": "taskkill communicate timeout"}
    so = (out_b or b"").decode("utf-8", errors="replace").strip()
    se = (err_b or b"").decode("utf-8", errors="replace").strip()
    rc = tk.returncode if tk.returncode is not None else -1
    low = se.lower()
    # 目标已退出时 taskkill 常非 0，仍视为可接受
    benign = rc == 0 or any(
        x in low
        for x in (
            "not found",
            "could not find",
            "no tasks",
            "no matching",
        )
    ) or ("没有运行" in se and "无法终止" not in se)
    return {
        "ok": benign,
        "returncode": rc,
        "stdout_tail": so[:800],
        "stderr_tail": se[:800],
    }


async def _terminate_subprocess(
    proc: asyncio.subprocess.Process, *, kill_subtree: bool = False
) -> None:
    """取消外部命令时尽量结束子进程；Windows 上可选整树结束（``task_shell`` 取消 GUI 用）。"""
    if proc.returncode is not None:
        return
    pid = proc.pid
    if sys.platform == "win32" and kill_subtree:
        await kill_windows_process_tree(pid)
    else:
        proc.terminate()
        try:
            await asyncio.wait_for(proc.wait(), timeout=5.0)
        except TimeoutError:
            proc.kill()
            try:
                await asyncio.wait_for(proc.wait(), timeout=2.0)
            except TimeoutError:
                pass
    if proc.returncode is None:
        try:
            proc.kill()
        except (ProcessLookupError, OSError):
            return
        try:
            await asyncio.wait_for(proc.wait(), timeout=2.0)
        except TimeoutError:
            pass


class ProcessTool:
    def __init__(self, policy: PathPolicy, cwd: Path | None = None) -> None:
        self._policy = policy
        self._cwd = cwd or Path.cwd()

    async def run_command(
        self,
        argv: list[str],
        *,
        cwd: str | None = None,
        env: dict[str, str] | None = None,
        timeout_sec: float | None = None,
        on_line: Callable[[str], Awaitable[None]] | None = None,
        on_spawned: Callable[[int], Awaitable[None]] | None = None,
        kill_subtree_on_cancel: bool | None = None,
    ) -> dict[str, Any]:
        wd = Path(cwd).resolve() if cwd else self._cwd
        self._policy.check_allowed(wd)
        proc = await asyncio.create_subprocess_exec(
            *argv,
            cwd=str(wd),
            env=subprocess_env_utf8(env),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        if on_spawned is not None:
            await on_spawned(proc.pid)
        subtree = (
            kill_subtree_on_cancel
            if kill_subtree_on_cancel is not None
            else (sys.platform == "win32")
        )
        out_chunks: list[str] = []
        err_chunks: list[str] = []

        async def pump(
            stream: asyncio.StreamReader | None, sink: list[str], prefix: str
        ) -> None:
            if stream is None:
                return
            while True:
                line = await stream.readline()
                if not line:
                    break
                s = line.decode("utf-8", errors="replace").rstrip("\r\n")
                sink.append(s)
                if on_line:
                    await on_line(f"{prefix}{s}")

        async def _pump_all() -> None:
            await asyncio.gather(
                pump(proc.stdout, out_chunks, "[stdout] "),
                pump(proc.stderr, err_chunks, "[stderr] "),
                proc.wait(),
            )

        try:
            if timeout_sec:
                await asyncio.wait_for(_pump_all(), timeout=timeout_sec)
            else:
                await _pump_all()
        except asyncio.CancelledError:
            await _terminate_subprocess(proc, kill_subtree=subtree)
            raise
        except TimeoutError:
            await _terminate_subprocess(proc, kill_subtree=subtree)
            return {
                "ok": False,
                "error": "timeout",
                "pid": proc.pid,
                "stdout": "\n".join(out_chunks[-200:]),
                "stderr": "\n".join(err_chunks[-200:]),
            }
        rc = proc.returncode if proc.returncode is not None else -1
        return {
            "ok": rc == 0,
            "returncode": rc,
            "pid": proc.pid,
            "stdout": "\n".join(out_chunks),
            "stderr": "\n".join(err_chunks),
        }



"""
AI 可调图像处理：缩放/裁剪/色彩/绘制/叠加/拼接及 OpenCV 常见算子。

Pillow 始终可用；OpenCV 算子需 ``pip install aicd[opencv]`` 或 ``opencv-python-headless``。
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable

import numpy as np
from PIL import Image, ImageDraw, ImageEnhance, ImageFont, ImageOps

MAX_IMAGE_BYTES = 45 * 1024 * 1024
MAX_PIXELS = 20_000_000  # ~20MP
MAX_SIDE = 16_384
MAX_CONCAT_IMAGES = 8


def _try_import_cv2():  # type: ignore[no-untyped-def]
    try:
        import cv2  # noqa: WPS433

        return cv2
    except ImportError:
        return None


def _check_size(w: int, h: int) -> str | None:
    if w < 1 or h < 1:
        return "invalid image dimensions"
    if w > MAX_SIDE or h > MAX_SIDE:
        return f"image side exceeds {MAX_SIDE}"
    if w * h > MAX_PIXELS:
        return f"image pixel count exceeds {MAX_PIXELS}"
    return None


def _load_rgba(path: Path) -> tuple[Image.Image, str | None]:
    sz = path.stat().st_size
    if sz > MAX_IMAGE_BYTES:
        return Image.new("RGBA", (1, 1)), f"file too large ({sz} bytes)"
    im = Image.open(path)
    im.load()
    err = _check_size(*im.size)
    if err:
        return Image.new("RGBA", (1, 1)), err
    return im.convert("RGBA"), None


def _save(im: Image.Image, out: Path) -> None:
    out.parent.mkdir(parents=True, exist_ok=True)
    suf = out.suffix.lower()
    if suf in (".jpg", ".jpeg"):
        rgb = Image.new("RGB", im.size, (255, 255, 255))
        rgb.paste(im, mask=im.split()[3] if im.mode == "RGBA" else None)
        rgb.save(out, format="JPEG", quality=92)
    else:
        im.save(out, format="PNG")


def _parse_rgba(spec: Any, default: tuple[int, int, int, int]) -> tuple[int, int, int, int]:
    if spec is None:
        return default
    if isinstance(spec, (list, tuple)) and len(spec) >= 3:
        r, g, b = int(spec[0]), int(spec[1]), int(spec[2])
        a = int(spec[3]) if len(spec) > 3 else 255
        return (max(0, min(255, r)), max(0, min(255, g)), max(0, min(255, b)), max(0, min(255, a)))
    return default


def _f(x: Any, default: float) -> float:
    try:
        return float(x)
    except (TypeError, ValueError):
        return default


def _i(x: Any, default: int) -> int:
    try:
        return int(x)
    except (TypeError, ValueError):
        return default


def run_image_transform(
    operation: str,
    input_path: Path,
    output_path: Path,
    *,
    params: dict[str, Any] | None,
    policy_check: Callable[[Path], None],
) -> dict[str, Any]:
    op = (operation or "").strip().lower().replace("-", "_")
    p = dict(params or {})
    if op in ("help", "list_operations", "operations_help"):
        return {
            "ok": True,
            "operations_help": operations_help(),
            "known_operations": list_image_operations(),
        }

    inp = input_path.expanduser().resolve()
    outp = output_path.expanduser().resolve()
    policy_check(outp)

    if op == "solid_canvas":
        try:
            out_im, meta = _op_solid_canvas(p)
        except ValueError as e:
            return {"ok": False, "error": str(e)}
        except Exception as e:
            return {"ok": False, "error": f"{type(e).__name__}: {e}"}
        e2 = _check_size(*out_im.size)
        if e2:
            return {"ok": False, "error": f"output {e2}"}
        try:
            _save(out_im, outp)
        except OSError as e:
            return {"ok": False, "error": str(e)}
        return {"ok": True, "path": str(outp), "operation": op, "meta": meta}

    policy_check(inp)
    if not inp.is_file():
        return {"ok": False, "error": f"input not found: {inp}"}

    im, err = _load_rgba(inp)
    if err:
        return {"ok": False, "error": err}

    try:
        if op == "resize":
            out_im, meta = _op_resize(im, p)
        elif op == "crop":
            out_im, meta = _op_crop(im, p)
        elif op == "rotate":
            out_im, meta = _op_rotate(im, p)
        elif op == "flip":
            out_im, meta = _op_flip(im, p)
        elif op == "brightness":
            out_im, meta = _op_enhance(im, p, "brightness")
        elif op == "contrast":
            out_im, meta = _op_enhance(im, p, "contrast")
        elif op == "saturation":
            out_im, meta = _op_enhance(im, p, "color")
        elif op == "sharpness":
            out_im, meta = _op_enhance(im, p, "sharpness")
        elif op == "grayscale":
            out_im, meta = _op_grayscale(im, p)
        elif op == "invert":
            out_im, meta = _op_invert(im, p)
        elif op == "posterize":
            out_im, meta = _op_posterize(im, p)
        elif op == "autocontrast":
            out_im, meta = _op_autocontrast(im, p)
        elif op == "equalize":
            out_im, meta = _op_equalize(im, p)
        elif op == "paste_overlay":
            out_im, meta = _op_paste_overlay(im, p, policy_check)
        elif op == "blend":
            out_im, meta = _op_blend(im, p, policy_check)
        elif op == "concat_horizontal":
            out_im, meta = _op_concat(im, p, policy_check, horizontal=True)
        elif op == "concat_vertical":
            out_im, meta = _op_concat(im, p, policy_check, horizontal=False)
        elif op == "draw_rectangle":
            out_im, meta = _op_draw_rect(im, p)
        elif op == "draw_ellipse":
            out_im, meta = _op_draw_ellipse(im, p)
        elif op == "draw_line":
            out_im, meta = _op_draw_line(im, p)
        elif op == "draw_polygon":
            out_im, meta = _op_draw_polygon(im, p)
        elif op == "draw_text":
            out_im, meta = _op_draw_text(im, p)
        elif op == "chroma_key":
            out_im, meta = _op_chroma_key(im, p)
        elif op == "tint":
            out_im, meta = _op_tint(im, p)
        elif op == "edge_canny":
            out_im, meta = _op_edge_canny(im, p)
        elif op == "gaussian_blur":
            out_im, meta = _op_gaussian_blur(im, p)
        elif op == "bilateral":
            out_im, meta = _op_bilateral(im, p)
        elif op == "morphology":
            out_im, meta = _op_morphology(im, p)
        elif op == "threshold":
            out_im, meta = _op_threshold(im, p)
        elif op == "laplacian":
            out_im, meta = _op_laplacian(im, p)
        elif op == "sobel":
            out_im, meta = _op_sobel(im, p)
        elif op == "grabcut_rect":
            out_im, meta = _op_grabcut_rect(im, p)
        else:
            return {
                "ok": False,
                "error": f"unknown operation: {operation!r}",
                "known_operations": list_image_operations(),
            }
    except ValueError as e:
        return {"ok": False, "error": str(e)}
    except Exception as e:
        return {"ok": False, "error": f"{type(e).__name__}: {e}"}

    e2 = _check_size(*out_im.size)
    if e2:
        return {"ok": False, "error": f"output {e2}"}
    _save(out_im, outp)
    return {"ok": True, "path": str(outp), "operation": op, "meta": meta}


def list_image_operations() -> list[str]:
    """公开给工具/测试：当前支持的操作名列表。"""
    return sorted(_known_ops())


def _known_ops() -> frozenset[str]:
    return frozenset(
        {
            "resize",
            "crop",
            "rotate",
            "flip",
            "brightness",
            "contrast",
            "saturation",
            "sharpness",
            "grayscale",
            "invert",
            "posterize",
            "autocontrast",
            "equalize",
            "paste_overlay",
            "blend",
            "concat_horizontal",
            "concat_vertical",
            "draw_rectangle",
            "draw_ellipse",
            "draw_line",
            "draw_polygon",
            "draw_text",
            "solid_canvas",
            "chroma_key",
            "tint",
            "edge_canny",
            "gaussian_blur",
            "bilateral",
            "morphology",
            "threshold",
            "laplacian",
            "sobel",
            "grabcut_rect",
        }
    )


def operations_help() -> dict[str, Any]:
    """供工具描述引用的简要参数说明。"""
    return {
        "resize": {
            "params": {
                "width": "int target width (optional if height set)",
                "height": "int target height",
                "mode": "fit | stretch | cover (default fit)",
            }
        },
        "crop": {"params": {"left": 0, "top": 0, "width": 100, "height": 100}},
        "rotate": {"params": {"degrees": 90, "expand": True, "fill": [255, 255, 255, 0]}},
        "flip": {"params": {"axis": "horizontal | vertical"}},
        "brightness": {"params": {"factor": 1.0, "note": "1.0 unchanged; try 0.5–2.0"}},
        "contrast": {"params": {"factor": 1.0}},
        "saturation": {"params": {"factor": 1.0}},
        "sharpness": {"params": {"factor": 1.0}},
        "grayscale": {},
        "invert": {},
        "posterize": {"params": {"bits": 4}},
        "autocontrast": {"params": {"cutoff": 0}},
        "equalize": {},
        "paste_overlay": {
            "params": {
                "overlay_path": "absolute path",
                "x": 0,
                "y": 0,
                "alpha": 1.0,
            }
        },
        "blend": {"params": {"second_path": "abs path", "alpha": 0.5}},
        "concat_horizontal": {
            "params": {
                "second_path": "optional",
                "paths": "optional list of abs paths; max 8 images total with input",
            }
        },
        "concat_vertical": {
            "params": {
                "second_path": "optional",
                "paths": "optional list of abs paths; max 8 images total with input",
            }
        },
        "draw_rectangle": {
            "params": {
                "xyxy": "[x0,y0,x1,y1]",
                "outline": "optional rgba",
                "fill": "optional rgba",
                "width": 2,
            }
        },
        "draw_ellipse": {
            "params": {"xyxy": "[x0,y0,x1,y1]", "outline": "optional rgba", "fill": "optional rgba"}
        },
        "draw_line": {
            "params": {"xyxy": "[x0,y0,x1,y1]", "width": 2, "fill": "[r,g,b,a]"}
        },
        "draw_polygon": {
            "params": {
                "points": "array of [x,y]",
                "outline": "optional rgba",
                "fill": "optional rgba",
            }
        },
        "draw_text": {
            "params": {
                "x": 0,
                "y": 0,
                "text": "str",
                "font_size": 16,
                "fill": [0, 0, 0, 255],
            }
        },
        "solid_canvas": {
            "params": {"width": 800, "height": 600, "fill": [255, 255, 255, 255]},
            "note": "ignores input image",
        },
        "chroma_key": {
            "params": {
                "lower_hsv": "[h,s,v] uint8 0-255",
                "upper_hsv": "[h,s,v]",
                "softness": 0,
            }
        },
        "tint": {"params": {"color": "[r,g,b]", "alpha": 0.3}},
        "opencv": {
            "edge_canny": {"low": 50, "high": 150},
            "gaussian_blur": {"ksize": 5, "sigma": 0},
            "bilateral": {"d": 9, "sigma_color": 75, "sigma_space": 75},
            "morphology": {"op": "erode|dilate|open|close", "ksize": 3, "iterations": 1},
            "threshold": {"type": "binary|otsu|adaptive", "thresh": 127, "maxval": 255},
            "laplacian": {"ksize": 3},
            "sobel": {"dx": 1, "dy": 0, "ksize": 3},
            "grabcut_rect": {"x": 0, "y": 0, "width": 100, "height": 100, "iterations": 5},
        },
    }


# --- Pillow ops ---


def _op_resize(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    w, h = im.size
    tw = _i(p.get("width"), 0)
    th = _i(p.get("height"), 0)
    mode = str(p.get("mode") or "fit").lower()
    if tw <= 0 and th <= 0:
        raise ValueError("resize requires width and/or height")
    if tw <= 0:
        tw = max(1, int(w * th / h))
    if th <= 0:
        th = max(1, int(h * tw / w))
    if mode == "stretch":
        out = im.resize((tw, th), Image.Resampling.LANCZOS)
    elif mode == "cover":
        scale = max(tw / w, th / h)
        nw, nh = max(1, int(w * scale)), max(1, int(h * scale))
        tmp = im.resize((nw, nh), Image.Resampling.LANCZOS)
        l = (nw - tw) // 2
        t = (nh - th) // 2
        out = tmp.crop((l, t, l + tw, t + th))
    else:  # fit
        scale = min(tw / w, th / h)
        nw, nh = max(1, int(w * scale)), max(1, int(h * scale))
        out = im.resize((nw, nh), Image.Resampling.LANCZOS)
    return out, {"size": out.size, "mode": mode}


def _op_crop(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    left = _i(p.get("left"), 0)
    top = _i(p.get("top"), 0)
    width = _i(p.get("width"), 0)
    height = _i(p.get("height"), 0)
    if width <= 0 or height <= 0:
        raise ValueError("crop requires positive width and height")
    box = (left, top, left + width, top + height)
    return im.crop(box), {"box": box}


def _op_rotate(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    deg = _f(p.get("degrees"), 0)
    expand = bool(p.get("expand", True))
    fill = _parse_rgba(p.get("fill"), (255, 255, 255, 0))
    out = im.rotate(
        -deg,
        resample=Image.Resampling.BICUBIC,
        expand=expand,
        fillcolor=fill,
    )
    return out, {"degrees": deg, "expand": expand}


def _op_flip(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    axis = str(p.get("axis") or "horizontal").lower()
    if axis == "vertical":
        return ImageOps.flip(im), {"axis": "vertical"}
    return ImageOps.mirror(im), {"axis": "horizontal"}


def _op_enhance(
    im: Image.Image, p: dict[str, Any], kind: str
) -> tuple[Image.Image, dict[str, Any]]:
    fac = _f(p.get("factor"), 1.0)
    if fac <= 0:
        raise ValueError("factor must be positive")
    base = im.convert("RGB")
    if kind == "brightness":
        e = ImageEnhance.Brightness(base)
    elif kind == "contrast":
        e = ImageEnhance.Contrast(base)
    elif kind == "color":
        e = ImageEnhance.Color(base)
    else:
        e = ImageEnhance.Sharpness(base)
    rgb = e.enhance(fac)
    r, g, b = rgb.split()
    a = im.split()[3]
    out = Image.merge("RGBA", (r, g, b, a))
    return out, {"factor": fac, "kind": kind}


def _op_grayscale(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    g = ImageOps.grayscale(im.convert("RGB"))
    a = im.split()[3]
    out = Image.merge("RGBA", (g, g, g, a))
    return out, {}


def _op_invert(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    rgb = ImageOps.invert(im.convert("RGB"))
    r, g, b = rgb.split()
    a = im.split()[3]
    return Image.merge("RGBA", (r, g, b, a)), {}


def _op_posterize(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    bits = max(1, min(8, _i(p.get("bits"), 4)))
    rgb = ImageOps.posterize(im.convert("RGB"), bits)
    r, g, b = rgb.split()
    a = im.split()[3]
    return Image.merge("RGBA", (r, g, b, a)), {"bits": bits}


def _op_autocontrast(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    co = max(0, min(50, _i(p.get("cutoff"), 0)))
    rgb = ImageOps.autocontrast(im.convert("RGB"), cutoff=co)
    r, g, b = rgb.split()
    a = im.split()[3]
    return Image.merge("RGBA", (r, g, b, a)), {"cutoff": co}


def _op_equalize(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    rgb = ImageOps.equalize(im.convert("RGB"))
    r, g, b = rgb.split()
    a = im.split()[3]
    return Image.merge("RGBA", (r, g, b, a)), {}


def _op_paste_overlay(
    im: Image.Image, p: dict[str, Any], policy_check: Callable[[Path], None]
) -> tuple[Image.Image, dict[str, Any]]:
    opath = p.get("overlay_path")
    if not opath:
        raise ValueError("paste_overlay requires params.overlay_path")
    ovp = Path(str(opath)).expanduser().resolve()
    policy_check(ovp)
    ov, err = _load_rgba(ovp)
    if err:
        raise ValueError(err)
    x, y = _i(p.get("x"), 0), _i(p.get("y"), 0)
    alpha = max(0.0, min(1.0, _f(p.get("alpha"), 1.0)))
    if alpha < 1.0:
        ov = ov.copy()
        _, _, _, a = ov.split()
        a = a.point(lambda v: int(v * alpha))
        ov.putalpha(a)
    canvas = im.copy()
    canvas.paste(ov, (x, y), ov)
    return canvas, {"overlay": str(ovp), "x": x, "y": y, "alpha": alpha}


def _op_blend(
    im: Image.Image, p: dict[str, Any], policy_check: Callable[[Path], None]
) -> tuple[Image.Image, dict[str, Any]]:
    sp = p.get("second_path")
    if not sp:
        raise ValueError("blend requires params.second_path")
    p2 = Path(str(sp)).expanduser().resolve()
    policy_check(p2)
    b, err = _load_rgba(p2)
    if err:
        raise ValueError(err)
    if b.size != im.size:
        b = b.resize(im.size, Image.Resampling.LANCZOS)
    a = max(0.0, min(1.0, _f(p.get("alpha"), 0.5)))
    out = Image.blend(im, b, a)
    return out, {"alpha": a, "second": str(p2)}


def _op_concat(
    im: Image.Image,
    p: dict[str, Any],
    policy_check: Callable[[Path], None],
    *,
    horizontal: bool,
) -> tuple[Image.Image, dict[str, Any]]:
    paths: list[Path] = []
    if p.get("second_path"):
        paths.append(Path(str(p["second_path"])).expanduser().resolve())
    extra = p.get("paths")
    if isinstance(extra, list):
        for x in extra:
            paths.append(Path(str(x)).expanduser().resolve())
    images: list[Image.Image] = [im]
    for path in paths:
        policy_check(path)
        o, err = _load_rgba(path)
        if err:
            raise ValueError(err)
        images.append(o)
    if len(images) > MAX_CONCAT_IMAGES:
        raise ValueError(f"at most {MAX_CONCAT_IMAGES} images")
    if horizontal:
        h = max(x.size[1] for x in images)
        resized = []
        for x in images:
            if x.size[1] != h:
                nw = max(1, int(x.size[0] * h / x.size[1]))
                resized.append(x.resize((nw, h), Image.Resampling.LANCZOS))
            else:
                resized.append(x)
        w = sum(x.size[0] for x in resized)
        out = Image.new("RGBA", (w, h), (0, 0, 0, 0))
        ox = 0
        for x in resized:
            out.paste(x, (ox, 0), x)
            ox += x.size[0]
    else:
        w = max(x.size[0] for x in images)
        resized = []
        for x in images:
            if x.size[0] != w:
                nh = max(1, int(x.size[1] * w / x.size[0]))
                resized.append(x.resize((w, nh), Image.Resampling.LANCZOS))
            else:
                resized.append(x)
        h = sum(x.size[1] for x in resized)
        out = Image.new("RGBA", (w, h), (0, 0, 0, 0))
        oy = 0
        for x in resized:
            out.paste(x, (0, oy), x)
            oy += x.size[1]
    return out, {"count": len(images), "horizontal": horizontal}


def _op_draw_rect(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    xy = p.get("xyxy")
    if not isinstance(xy, (list, tuple)) or len(xy) < 4:
        raise ValueError("draw_rectangle needs params.xyxy [x0,y0,x1,y1]")
    x0, y0, x1, y1 = int(xy[0]), int(xy[1]), int(xy[2]), int(xy[3])
    outline = p.get("outline")
    fill = p.get("fill")
    width = max(1, _i(p.get("width"), 2))
    canvas = im.copy()
    dr = ImageDraw.Draw(canvas, "RGBA")
    kw: dict[str, Any] = {"width": width}
    if outline is not None:
        kw["outline"] = _parse_rgba(outline, (0, 0, 0, 255))
    if fill is not None:
        kw["fill"] = _parse_rgba(fill, (255, 0, 0, 128))
    dr.rectangle([x0, y0, x1, y1], **kw)
    return canvas, {"xyxy": [x0, y0, x1, y1]}


def _op_draw_ellipse(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    xy = p.get("xyxy")
    if not isinstance(xy, (list, tuple)) or len(xy) < 4:
        raise ValueError("draw_ellipse needs params.xyxy")
    x0, y0, x1, y1 = int(xy[0]), int(xy[1]), int(xy[2]), int(xy[3])
    width = max(1, _i(p.get("width"), 2))
    canvas = im.copy()
    dr = ImageDraw.Draw(canvas, "RGBA")
    kw: dict[str, Any] = {"width": width}
    if p.get("outline") is not None:
        kw["outline"] = _parse_rgba(p["outline"], (0, 0, 0, 255))
    if p.get("fill") is not None:
        kw["fill"] = _parse_rgba(p["fill"], (255, 0, 0, 80))
    dr.ellipse([x0, y0, x1, y1], **kw)
    return canvas, {}


def _op_draw_line(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    xy = p.get("xyxy")
    if not isinstance(xy, (list, tuple)) or len(xy) < 4:
        raise ValueError("draw_line needs params.xyxy")
    width = max(1, _i(p.get("width"), 2))
    fill = _parse_rgba(p.get("fill"), (0, 0, 0, 255))
    canvas = im.copy()
    dr = ImageDraw.Draw(canvas, "RGBA")
    dr.line(
        [(int(xy[0]), int(xy[1])), (int(xy[2]), int(xy[3]))],
        fill=fill,
        width=width,
    )
    return canvas, {}


def _op_draw_polygon(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    pts = p.get("points")
    if not isinstance(pts, list) or len(pts) < 2:
        raise ValueError("draw_polygon needs params.points [[x,y],...]")
    flat = []
    for q in pts:
        if isinstance(q, (list, tuple)) and len(q) >= 2:
            flat.append((int(q[0]), int(q[1])))
    if len(flat) < 2:
        raise ValueError("invalid points")
    canvas = im.copy()
    dr = ImageDraw.Draw(canvas, "RGBA")
    kw: dict[str, Any] = {}
    if p.get("outline") is not None:
        kw["outline"] = _parse_rgba(p["outline"], (0, 0, 0, 255))
    if p.get("fill") is not None:
        kw["fill"] = _parse_rgba(p["fill"], (0, 128, 255, 100))
    dr.polygon(flat, **kw)
    return canvas, {"vertices": len(flat)}


def _op_draw_text(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    text = str(p.get("text") or "")
    if not text:
        raise ValueError("draw_text needs params.text")
    x, y = _i(p.get("x"), 0), _i(p.get("y"), 0)
    fs = max(8, min(256, _i(p.get("font_size"), 16)))
    fill = _parse_rgba(p.get("fill"), (0, 0, 0, 255))
    canvas = im.copy()
    dr = ImageDraw.Draw(canvas, "RGBA")
    try:
        font = ImageFont.truetype("arial.ttf", fs)
    except OSError:
        try:
            font = ImageFont.truetype("DejaVuSans.ttf", fs)
        except OSError:
            font = ImageFont.load_default()
    dr.text((x, y), text, fill=fill, font=font)
    return canvas, {"chars": len(text)}


def _op_solid_canvas(p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    w0, h0 = _i(p.get("width"), 0), _i(p.get("height"), 0)
    if w0 <= 0 or h0 <= 0:
        raise ValueError("solid_canvas requires positive params.width and params.height")
    w = max(1, min(MAX_SIDE, w0))
    h = max(1, min(MAX_SIDE, h0))
    if w * h > MAX_PIXELS:
        raise ValueError("canvas exceeds MAX_PIXELS")
    fill = _parse_rgba(p.get("fill"), (255, 255, 255, 255))
    return Image.new("RGBA", (w, h), fill), {"size": (w, h)}


def _op_chroma_key(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    lo = p.get("lower_hsv")
    hi = p.get("upper_hsv")
    if not isinstance(lo, (list, tuple)) or not isinstance(hi, (list, tuple)):
        raise ValueError("chroma_key needs lower_hsv and upper_hsv [h,s,v] 0-255 uint8 style")
    rgb = im.convert("RGB")
    hsv = rgb.convert("HSV")
    arr = np.array(hsv)
    loa = np.array([int(lo[0]), int(lo[1]), int(lo[2])], dtype=np.uint8)
    hia = np.array([int(hi[0]), int(hi[1]), int(hi[2])], dtype=np.uint8)
    mask = np.all((arr >= loa) & (arr <= hia), axis=-1)
    a = np.array(im.split()[3])
    a[mask] = 0
    out = Image.merge("RGBA", (*rgb.split(), Image.fromarray(a)))
    return out, {"masked_pixels": int(mask.sum())}


def _op_tint(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    c = p.get("color")
    if not isinstance(c, (list, tuple)) or len(c) < 3:
        raise ValueError("tint needs params.color [r,g,b]")
    alpha = max(0.0, min(1.0, _f(p.get("alpha"), 0.3)))
    tint = Image.new("RGBA", im.size, (int(c[0]), int(c[1]), int(c[2]), int(255 * alpha)))
    return Image.alpha_composite(im, tint), {"alpha": alpha}


def _pil_to_cv_bgr(im: Image.Image):  # type: ignore[no-untyped-def]
    import numpy as np

    rgb = np.array(im.convert("RGB"))
    cv2 = _try_import_cv2()
    if cv2 is None:
        raise ValueError("OpenCV not installed; pip install opencv-python-headless")
    return cv2.cvtColor(rgb, cv2.COLOR_RGB2BGR)


def _cv_bgr_to_pil_rgba(bgr, alpha: Image.Image | None = None) -> Image.Image:  # type: ignore[no-untyped-def]
    import numpy as np

    cv2 = _try_import_cv2()
    assert cv2 is not None
    rgb = cv2.cvtColor(bgr, cv2.COLOR_BGR2RGB)
    pil = Image.fromarray(rgb)
    if alpha is not None and alpha.size == pil.size:
        return Image.merge("RGBA", (*pil.split(), alpha))
    return pil.convert("RGBA")


def _op_edge_canny(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    cv2 = _try_import_cv2()
    if cv2 is None:
        raise ValueError("edge_canny requires OpenCV (pip install opencv-python-headless)")
    low = max(0, _i(p.get("low"), 50))
    high = max(0, _i(p.get("high"), 150))
    bgr = _pil_to_cv_bgr(im)
    gray = cv2.cvtColor(bgr, cv2.COLOR_BGR2GRAY)
    edges = cv2.Canny(gray, low, high)
    edges_rgba = cv2.cvtColor(edges, cv2.COLOR_GRAY2BGRA)
    edges_rgba[:, :, 3] = 255
    rgb = cv2.cvtColor(edges_rgba, cv2.COLOR_BGRA2RGBA)
    return Image.fromarray(rgb), {"low": low, "high": high}


def _op_gaussian_blur(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    cv2 = _try_import_cv2()
    if cv2 is None:
        raise ValueError("gaussian_blur requires OpenCV")
    k = max(1, _i(p.get("ksize"), 5)) | 1  # odd
    sigma = _f(p.get("sigma"), 0)
    bgr = _pil_to_cv_bgr(im)
    blur = cv2.GaussianBlur(bgr, (k, k), sigma)
    a = im.split()[3]
    return _cv_bgr_to_pil_rgba(blur, a), {"ksize": k}


def _op_bilateral(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    cv2 = _try_import_cv2()
    if cv2 is None:
        raise ValueError("bilateral requires OpenCV")
    d = max(1, _i(p.get("d"), 9))
    sc = _f(p.get("sigma_color"), 75)
    ss = _f(p.get("sigma_space"), 75)
    bgr = _pil_to_cv_bgr(im)
    out = cv2.bilateralFilter(bgr, d, sc, ss)
    a = im.split()[3]
    return _cv_bgr_to_pil_rgba(out, a), {}


def _op_morphology(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    cv2 = _try_import_cv2()
    if cv2 is None:
        raise ValueError("morphology requires OpenCV")
    opn = str(p.get("op") or "close").lower()
    ks = max(1, _i(p.get("ksize"), 3)) | 1
    it = max(1, _i(p.get("iterations"), 1))
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (ks, ks))
    gray = cv2.cvtColor(_pil_to_cv_bgr(im), cv2.COLOR_BGR2GRAY)
    _, bw = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    if opn == "erode":
        proc = cv2.erode(bw, kernel, iterations=it)
    elif opn == "dilate":
        proc = cv2.dilate(bw, kernel, iterations=it)
    elif opn == "open":
        proc = cv2.morphologyEx(bw, cv2.MORPH_OPEN, kernel, iterations=it)
    else:
        proc = cv2.morphologyEx(bw, cv2.MORPH_CLOSE, kernel, iterations=it)
    bgra = cv2.cvtColor(proc, cv2.COLOR_GRAY2BGRA)
    bgra[:, :, 3] = 255
    return Image.fromarray(cv2.cvtColor(bgra, cv2.COLOR_BGRA2RGBA)), {"op": opn}


def _op_threshold(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    cv2 = _try_import_cv2()
    if cv2 is None:
        raise ValueError("threshold requires OpenCV")
    typ = str(p.get("type") or "binary").lower()
    gray = cv2.cvtColor(_pil_to_cv_bgr(im), cv2.COLOR_BGR2GRAY)
    maxv = _i(p.get("maxval"), 255)
    if typ == "otsu":
        _, bw = cv2.threshold(gray, 0, maxv, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    elif typ == "adaptive":
        bw = cv2.adaptiveThreshold(
            gray, maxv, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 11, 2
        )
    else:
        th = _i(p.get("thresh"), 127)
        _, bw = cv2.threshold(gray, th, maxv, cv2.THRESH_BINARY)
    rgba = cv2.cvtColor(bw, cv2.COLOR_GRAY2RGBA)
    return Image.fromarray(rgba), {"type": typ}


def _op_laplacian(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    cv2 = _try_import_cv2()
    if cv2 is None:
        raise ValueError("laplacian requires OpenCV")
    ks = max(1, _i(p.get("ksize"), 3)) | 1
    gray = cv2.cvtColor(_pil_to_cv_bgr(im), cv2.COLOR_BGR2GRAY)
    lap = cv2.Laplacian(gray, cv2.CV_16S, ksize=ks)
    lap = cv2.convertScaleAbs(lap)
    rgba = cv2.cvtColor(lap, cv2.COLOR_GRAY2RGBA)
    return Image.fromarray(rgba), {"ksize": ks}


def _op_sobel(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    cv2 = _try_import_cv2()
    if cv2 is None:
        raise ValueError("sobel requires OpenCV")
    dx = max(0, min(2, _i(p.get("dx"), 1)))
    dy = max(0, min(2, _i(p.get("dy"), 0)))
    ks = max(1, _i(p.get("ksize"), 3)) | 1
    gray = cv2.cvtColor(_pil_to_cv_bgr(im), cv2.COLOR_BGR2GRAY)
    s = cv2.Sobel(gray, cv2.CV_16S, dx, dy, ksize=ks)
    s = cv2.convertScaleAbs(s)
    rgba = cv2.cvtColor(s, cv2.COLOR_GRAY2RGBA)
    return Image.fromarray(rgba), {"dx": dx, "dy": dy}


def _op_grabcut_rect(im: Image.Image, p: dict[str, Any]) -> tuple[Image.Image, dict[str, Any]]:
    cv2 = _try_import_cv2()
    if cv2 is None:
        raise ValueError("grabcut_rect requires OpenCV")
    x, y = _i(p.get("x"), 0), _i(p.get("y"), 0)
    w, h = _i(p.get("width"), 0), _i(p.get("height"), 0)
    if w <= 0 or h <= 0:
        raise ValueError("grabcut_rect needs positive width/height")
    it = max(1, min(10, _i(p.get("iterations"), 5)))
    bgr = _pil_to_cv_bgr(im)
    H0, W0 = bgr.shape[0], bgr.shape[1]
    x = max(0, min(W0 - 1, x))
    y = max(0, min(H0 - 1, y))
    w = max(1, min(W0 - x, w))
    h = max(1, min(H0 - y, h))
    mask = np.zeros(bgr.shape[:2], np.uint8)
    bgd = np.zeros((1, 65), np.float64)
    fgd = np.zeros((1, 65), np.float64)
    rect = (x, y, w, h)
    cv2.grabCut(bgr, mask, rect, bgd, fgd, it, cv2.GC_INIT_WITH_RECT)
    mask2 = np.where((mask == 2) | (mask == 0), 0, 255).astype("uint8")  # type: ignore[name-defined]
    rgba = cv2.cvtColor(bgr, cv2.COLOR_BGR2BGRA)
    rgba[:, :, 3] = mask2
    pil = Image.fromarray(cv2.cvtColor(rgba, cv2.COLOR_BGRA2RGBA))
    return pil, {"rect": list(rect), "iterations": it}



"""Windows / Unix shell argv 构造。"""

from __future__ import annotations

import sys

import pytest

from aicd.tools import process_tool as pt


def test_shell_argv_posix(monkeypatch) -> None:
    monkeypatch.setattr(sys, "platform", "linux")
    av = pt.shell_argv_from_user_input(command="echo hi")
    assert av == ["/bin/sh", "-c", "echo hi"]


def test_shell_argv_posix_direct_argv(monkeypatch) -> None:
    monkeypatch.setattr(sys, "platform", "linux")
    av = pt.shell_argv_from_user_input(argv=["/bin/echo", "a b"])
    assert av == ["/bin/echo", "a b"]


def test_shell_argv_windows_cmd(monkeypatch) -> None:
    monkeypatch.setattr(sys, "platform", "win32")
    monkeypatch.setenv("ComSpec", r"C:\Windows\System32\cmd.exe")
    av = pt.shell_argv_from_user_input(command="dir")
    assert av[0].lower().endswith("cmd.exe")
    assert av[1] == "/d"
    assert av[2] == "/c"
    assert "chcp 65001" in av[3]
    assert av[3].endswith("& dir") or "& dir" in av[3]


def test_shell_argv_windows_powershell(monkeypatch) -> None:
    monkeypatch.setattr(sys, "platform", "win32")
    av = pt.shell_argv_from_user_input(command="Get-Location", shell="powershell")
    assert av[0].lower().endswith("powershell.exe") or "powershell" in av[0].lower()
    assert "-Command" in av
    assert av[-1] == "Get-Location"


def test_shell_argv_errors() -> None:
    with pytest.raises(ValueError, match="non-empty"):
        pt.shell_argv_from_user_input(command="   ")
    with pytest.raises(ValueError, match="argv"):
        pt.shell_argv_from_user_input(argv=[])


def test_default_shell_argv_alias(monkeypatch) -> None:
    monkeypatch.setattr(sys, "platform", "linux")
    assert pt.default_shell_argv("x") == ["/bin/sh", "-c", "x"]



from __future__ import annotations

import asyncio
from pathlib import Path


def test_build_runtime(tmp_path: Path) -> None:
    from aicd.runtime import build_runtime

    async def _go() -> None:
        ctx = await build_runtime(home=tmp_path / "h")
        assert ctx.home == (tmp_path / "h").resolve()
        rows = await ctx.tasks.list_tasks(10)
        assert isinstance(rows, list)
        await ctx.conn.close()

    asyncio.run(_go())


def test_fs_list_self(tmp_path: Path) -> None:
    from aicd.config import AppConfig
    from aicd.tools.fs_tool import FSTool
    from aicd.tools.path_policy import PathPolicy

    cfg = AppConfig()
    pol = PathPolicy(cfg.paths)
    fs = FSTool(pol, cwd=tmp_path)
    (tmp_path / "a.txt").write_text("hi", encoding="utf-8")
    r = fs.list_dir(".")
    assert r["ok"] is True


def test_code_outline() -> None:
    from aicd.tools.code_intel import python_outline

    r = python_outline(__file__)
    assert r.get("ok") is True



from __future__ import annotations

import asyncio
import json
import os
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any

import aiosqlite

from aicd.agent.chat_context import (
    build_hydrated_messages,
    longest_valid_openai_suffix,
    repair_openai_chat_messages,
)
from aicd.agent.chat_serialize import tool_calls_to_jsonable
from aicd.agent.llm import LLMClient
from aicd.agent.idle_pulse import sub_agent_idle_pulse_user_text
from aicd.agent.prompts import MAIN_SYSTEM, SUB_SYSTEM
from aicd.env_probe import RUNTIME_ENV_FILENAME
from aicd.agent.tool_router import ToolRouter
from aicd.agent.tools_openai import openai_tools
from aicd.bus.event_bus import AsyncEventBus
from aicd.config import AppConfig
from aicd.db.store import TaskStore
from aicd.runtime_task_context import current_task_id
from aicd.tasks.manager import TaskManager

if TYPE_CHECKING:
    from aicd.session_log import SessionLogger


def _trim_chat_messages_to_last_user_turn(messages: list[dict[str, Any]]) -> None:
    """
    原地裁剪：只保留「最后一条 user」及其后的消息。

    ``chat_new_session`` 会切换 DB session_id；若在工具循环中途 ``messages.clear()``，
    会删掉带 ``tool_calls`` 的 assistant，只留下孤立的 ``role=tool``，触发 API 400。
    因此在整轮 tool 结果都追加后，再删掉更早轮次，保留 ``user → assistant(tool_calls) → tool*`` 链。
    """
    last_u = -1
    for i, m in enumerate(messages):
        if m.get("role") == "user":
            last_u = i
    if last_u > 0:
        del messages[:last_u]


def _sub_agent_cancel_requested() -> bool:
    ct = asyncio.current_task()
    if ct is None:
        return False
    fn = getattr(ct, "cancelling", None)
    return callable(fn) and bool(fn())


def _format_task_inbox_block(rows: list[Any]) -> str:
    parts: list[str] = []
    for m in rows:
        body = (
            json.dumps(m.body, ensure_ascii=False)
            if getattr(m, "body", None) is not None
            else "{}"
        )
        parts.append(
            f"[task_message id={m.id} from={m.from_task_id} kind={m.kind}]\n{body}"
        )
    return "\n\n".join(parts)


def _trim_sub_agent_messages(messages: list[dict[str, Any]], *, max_msgs: int) -> None:
    """
    子 Agent 上下文过长时裁剪为最近若干条。

    禁止简单 ``tail = messages[-k:]``：可能切断 ``assistant(tool_calls)`` 与其 ``tool`` 结果，
    导致 DashScope/OpenAI 兼容接口报
    ``must be a \"tool\" message``。改为取**合法**的最长后缀。
    """
    if not messages:
        return
    sys0 = messages[0]
    body = messages[1:]
    max_body = max(0, max_msgs - 1)
    if max_body <= 0:
        messages.clear()
        messages.append(sys0)
        return
    if len(body) <= max_body:
        repair_openai_chat_messages(body)
        messages.clear()
        messages.extend([sys0, *body])
        return
    picked = longest_valid_openai_suffix(body, max_body)
    repair_openai_chat_messages(picked)
    messages.clear()
    messages.extend([sys0, *picked])


def _msg_assistant_dict(msg: Any) -> dict[str, Any]:
    out: dict[str, Any] = {
        "role": "assistant",
        "content": msg.content,
    }
    if msg.tool_calls:
        out["tool_calls"] = [
            {
                "id": tc.id,
                "type": "function",
                "function": {
                    "name": tc.function.name,
                    "arguments": tc.function.arguments or "{}",
                },
            }
            for tc in msg.tool_calls
        ]
    return out


class AgentLoop:
    def __init__(
        self,
        cfg: AppConfig,
        llm: LLMClient,
        router: ToolRouter,
        task_manager: TaskManager,
        store: TaskStore,
        conn: aiosqlite.Connection,
        layout: dict[str, Path],
        bus: AsyncEventBus,
        session_id: str,
        session_log: SessionLogger | None = None,
        env_capabilities_block: str = "",
    ) -> None:
        self._cfg = cfg
        self._llm = llm
        self._router = router
        self._tasks = task_manager
        self._store = store
        self._conn = conn
        self._layout = layout
        self._bus = bus
        self._session_id = session_id
        self._session_log = session_log
        self._env_capabilities_block = (env_capabilities_block or "").strip()
        self._messages: list[dict[str, Any]] = []
        self._strip_chat_prefix_after_tool_round = False
        self._workspace_path: Path = router.cwd
        o, t = router.get_ai_paths()
        self._ai_out: Path = o
        self._ai_tmp: Path = t

    def apply_chat_session_id(self, new_session_id: str) -> None:
        """切换主会话键；新消息写入新 session_id（旧记录在 DB 中仍保留）。

        不在此处清空 ``_messages``：工具 ``chat_new_session`` 可能在同一轮 assistant
        ``tool_calls`` 执行过程中被调用，立即清空会导致下一条请求只有孤立的 ``tool`` 角色。
        改在本轮所有 tool 结果写入后，再裁剪到「最后一条 user」起的后缀。
        """
        self._session_id = new_session_id
        self._strip_chat_prefix_after_tool_round = True

    def _log_llm_usage(self, resp: Any) -> None:
        slog = self._session_log
        if not slog:
            return
        u = getattr(resp, "usage", None)
        if u is not None:
            slog.write_event("llm", "LLM", "chat.completions", "", str(u))

    async def _prepend_main_inbox_if_any(self) -> None:
        """主对话：将 main-inbox-<session> 中未投递的 task_message 注入为 user（并 mark read）。"""
        from aicd.tasks.main_inbox import main_inbox_id

        try:
            to_id = main_inbox_id(self._session_id)
        except ValueError:
            return
        rows = await self._store.list_task_messages_unread(
            self._conn,
            to_task_id=to_id,
            limit=50,
        )
        if not rows:
            return
        block = _format_task_inbox_block(rows)
        await self._store.mark_task_messages_read(self._conn, [m.id for m in rows])
        self._messages.append(
            {
                "role": "user",
                "content": (
                    "【主收件箱】以下为未读消息（含子任务 **lifecycle** 与其它上行 **task_message**）。"
                    "请视情况 **task_result**、**task_session_children_digest** 或发起新动作。\n\n"
                    + block
                ),
            },
        )

    def set_workspace(self, path: Path) -> None:
        """与 ToolRouter.set_cwd 同步，用于系统提示中的工作目录说明。"""
        self._workspace_path = path.resolve()

    def set_ai_storage_paths(self, output: Path, tmp: Path) -> None:
        self._ai_out = output.resolve()
        self._ai_tmp = tmp.resolve()

    def _runtime_env_file_str(self) -> str:
        return str((self._layout["data"] / RUNTIME_ENV_FILENAME).resolve())

    async def hydrate_chat_from_store(self) -> None:
        """从 SQLite 恢复本会话消息（按条数尾部 + token 预算 + 可选摘记）。"""
        total = await self._store.count_chat_messages(self._conn, self._session_id)
        if total == 0:
            self._messages = []
            return
        max_m = self._cfg.agent.chat_history_max_messages
        truncated = total > max_m
        tail = min(total, max_m)
        offset = max(0, total - tail)
        rows = await self._store.list_chat_messages(
            self._conn, self._session_id, limit=tail, offset=offset
        )
        latest_summary: str | None = None
        if truncated:
            sums = await self._store.list_chat_summaries(
                self._conn, self._session_id, limit=50
            )
            if sums:
                latest_summary = sums[-1].content
        self._messages = build_hydrated_messages(
            rows=rows,
            truncated_by_count=truncated,
            latest_summary=latest_summary,
            context_budget_tokens=self._cfg.agent.context_budget_tokens,
        )

    def _system_content(self) -> str:
        base = (
            f"{MAIN_SYSTEM}\n\n"
            f"AI workspace: {self._workspace_path}\n"
            f"AI output directory: {self._ai_out}\n"
            f"AI tmp (scratch/intermediate, self-maintain): {self._ai_tmp}\n"
            f"Main chat session_id: {self._session_id}\n"
            f"Runtime env file: {self._runtime_env_file_str()}"
        )
        if self._env_capabilities_block:
            base = f"{base}\n\n{self._env_capabilities_block}"
        return base

    async def run_sub_agent_worker(self, task_id: str, instruction: str) -> None:
        slog = self._session_log
        if slog:
            ins = instruction if len(instruction) <= 8000 else instruction[:7997] + "..."
            slog.write_event(
                "task",
                "AgentLoop",
                "sub_agent_start",
                f"task_id={task_id}",
                ins,
            )
        ws = self._layout["tasks"] / task_id / "workspace"
        ws.mkdir(parents=True, exist_ok=True)
        out, tmp = self._router.get_ai_paths()
        row = await self._store.get_task(self._conn, task_id)
        payload: dict[str, Any] = row.payload if row else {}
        scratch_s = payload.get("scratch_dir")
        sub_tmp = (
            Path(scratch_s).resolve()
            if isinstance(scratch_s, str) and scratch_s.strip()
            else (tmp / "tasks" / task_id).resolve()
        )
        sub_tmp.mkdir(parents=True, exist_ok=True)
        token = current_task_id.set(task_id)
        try:
            sub = ToolRouter(
                self._cfg,
                layout=self._layout,
                task_manager=self._tasks,
                cwd=ws,
                session_log=slog,
                chat_session_id=None,
            )
            sub.set_ai_paths(out, sub_tmp)
            sub.set_vision_llm(self._llm)
            sub_sys = SUB_SYSTEM.format(
                cwd=str(ws),
                ai_out=str(out),
                ai_tmp=str(sub_tmp),
                runtime_env_file=self._runtime_env_file_str(),
            )
            extras: list[str] = []
            dd = payload.get("deliverable_dir")
            if isinstance(dd, str) and dd.strip():
                extras.append(
                    f"User-visible deliverable directory (prefer final artifacts here): {dd.strip()}"
                )
            ro = payload.get("read_only_paths")
            if isinstance(ro, list) and ro:
                paths = [str(x) for x in ro if isinstance(x, str) and x.strip()]
                if paths:
                    extras.append(
                        "Read-only reference paths (do not modify; you may read): "
                        + "; ".join(paths[:20])
                    )
            he = payload.get("parent_handoff_expectation")
            if isinstance(he, str) and he.strip():
                extras.append(f"Parent handoff expectation: {he.strip()[:1200]}")
            if extras:
                sub_sys = f"{sub_sys}\n\n" + "\n".join(extras)
            if self._env_capabilities_block:
                sub_sys = f"{sub_sys}\n\n{self._env_capabilities_block}"
            messages: list[dict[str, Any]] = [
                {"role": "system", "content": sub_sys},
                {"role": "user", "content": instruction},
            ]
            tools = openai_tools()
            max_r = self._cfg.agent.sub_agent_max_tool_rounds
            total_cap = max(300, max_r * 6)
            total_rounds = 0
            last_inbox_id = 0
            max_ctx_msgs = 100

            while True:
                if _sub_agent_cancel_requested():
                    raise asyncio.CancelledError
                rounds_left = min(max_r, total_cap - total_rounds)
                if rounds_left <= 0:
                    if slog:
                        slog.write_event(
                            "task",
                            "AgentLoop",
                            "sub_agent_total_cap",
                            f"task_id={task_id}",
                            f"cumulative={total_rounds} cap={total_cap}",
                        )
                    await self._bus.publish(
                        {
                            "topic": "task",
                            "task_id": task_id,
                            "event": "log",
                            "line": f"[ai] 子 Agent 已达累计 LLM 轮次上限 {total_cap}，结束",
                        }
                    )
                    return

                got_final = False
                for round_i in range(rounds_left):
                    if _sub_agent_cancel_requested():
                        raise asyncio.CancelledError
                    total_rounds += 1
                    burst_idx = round_i + 1
                    _trim_sub_agent_messages(messages, max_msgs=max_ctx_msgs)
                    await self._tasks.patch_task_payload(
                        task_id,
                        {
                            "sub_agent_trace": {
                                "phase": "llm_turn",
                                "llm_round": burst_idx,
                                "max_llm_rounds": max_r,
                                "cumulative_turns": total_rounds,
                                "cumulative_cap": total_cap,
                                "updated_at": datetime.now(timezone.utc).isoformat(),
                            }
                        },
                    )
                    await self._bus.publish(
                        {
                            "topic": "task",
                            "task_id": task_id,
                            "event": "progress",
                            "phase": "sub_agent",
                            "llm_round": burst_idx,
                            "max_llm_rounds": max_r,
                            "message": (
                                f"子 Agent：本轮第 {burst_idx}/{rounds_left} 次 LLM"
                                f"（累计 {total_rounds}/{total_cap}）"
                            ),
                        }
                    )
                    resp = await self._llm.chat(messages, tools)
                    self._log_llm_usage(resp)
                    choice = resp.choices[0]
                    msg = choice.message
                    if not msg.tool_calls:
                        final = msg.content or ""
                        messages.append({"role": "assistant", "content": final})
                        await self._bus.publish(
                            {
                                "topic": "task",
                                "task_id": task_id,
                                "event": "log",
                                "line": f"[ai] {final[:2000]}",
                            }
                        )
                        if slog:
                            slog.write_event(
                                "task",
                                "AgentLoop",
                                "sub_agent_reply",
                                f"task_id={task_id}",
                                final if len(final) <= 12000 else final[:11997] + "...",
                            )
                        got_final = True
                        break
                    messages.append(_msg_assistant_dict(msg))
                    tool_names: list[str] = []
                    for tc in msg.tool_calls:
                        if _sub_agent_cancel_requested():
                            raise asyncio.CancelledError
                        name = tc.function.name
                        tool_names.append(name)
                        raw_args = tc.function.arguments or "{}"
                        result = await sub.run(name, raw_args)
                        line = f"tool {name} -> {str(result)[:800]}"
                        await self._bus.publish(
                            {
                                "topic": "task",
                                "task_id": task_id,
                                "event": "log",
                                "line": line,
                            }
                        )
                        messages.append(
                            {
                                "role": "tool",
                                "tool_call_id": tc.id,
                                "content": json.dumps(result, ensure_ascii=False)[
                                    :24_000
                                ],
                            }
                        )
                    last_tool = tool_names[-1] if tool_names else ""
                    await self._tasks.patch_task_payload(
                        task_id,
                        {
                            "sub_agent_trace": {
                                "phase": "after_tools",
                                "llm_round": burst_idx,
                                "max_llm_rounds": max_r,
                                "cumulative_turns": total_rounds,
                                "tools_in_round": tool_names,
                                "last_tool": last_tool,
                                "updated_at": datetime.now(timezone.utc).isoformat(),
                            }
                        },
                    )
                    await self._bus.publish(
                        {
                            "topic": "task",
                            "task_id": task_id,
                            "event": "progress",
                            "phase": "sub_agent",
                            "llm_round": burst_idx,
                            "max_llm_rounds": max_r,
                            "last_tool": last_tool,
                            "tools": ",".join(tool_names[:12]),
                            "message": (
                                f"子 Agent：本轮 {burst_idx}/{rounds_left} 已执行 "
                                f"{len(tool_names)} 个工具"
                                + (f"（末次 {last_tool}）" if last_tool else "")
                            ),
                        }
                    )

                if not got_final and slog:
                    slog.write_event(
                        "task",
                        "AgentLoop",
                        "sub_agent_burst_exhausted",
                        f"task_id={task_id}",
                        f"burst_rounds={rounds_left}",
                    )

                await asyncio.sleep(0.12)
                inbox_rows = await self._store.list_task_messages(
                    self._conn,
                    to_task_id=task_id,
                    since_id=last_inbox_id,
                    limit=50,
                )
                hb_sub = self._cfg.agent.sub_agent_idle_heartbeat_sec
                if not inbox_rows and hb_sub > 0:
                    children = await self._store.list_tasks_by_parent(
                        self._conn, task_id
                    )
                    if any(r.status in ("running", "pending") for r in children):
                        sleep_s = max(5, min(3600, hb_sub))
                        await asyncio.sleep(sleep_s)
                        inbox_rows = await self._store.list_task_messages(
                            self._conn,
                            to_task_id=task_id,
                            since_id=last_inbox_id,
                            limit=50,
                        )

                if inbox_rows:
                    last_inbox_id = max(m.id for m in inbox_rows)
                    block = _format_task_inbox_block(inbox_rows)
                    messages.append(
                        {
                            "role": "user",
                            "content": (
                                "以下是通过 **task_message** 送达的新指令/补充"
                                "（上级主 AI 或并列任务）。请据此继续执行；若与当前工作冲突，"
                                "以最新指令为准并简要说明取舍。\n\n"
                                + block
                            ),
                        }
                    )
                    await self._store.mark_task_messages_read(
                        self._conn, [m.id for m in inbox_rows]
                    )
                    await self._bus.publish(
                        {
                            "topic": "task",
                            "task_id": task_id,
                            "event": "log",
                            "line": f"[inbox] 收到 {len(inbox_rows)} 条消息，继续执行",
                        }
                    )
                    await self._tasks.patch_task_payload(
                        task_id,
                        {
                            "sub_agent_trace": {
                                "phase": "inbox_continue",
                                "last_inbox_id": last_inbox_id,
                                "updated_at": datetime.now(timezone.utc).isoformat(),
                            }
                        },
                    )
                    continue

                if hb_sub <= 0:
                    return
                children2 = await self._store.list_tasks_by_parent(
                    self._conn, task_id
                )
                if not any(r.status in ("running", "pending") for r in children2):
                    return
                messages.append(
                    {
                        "role": "user",
                        "content": sub_agent_idle_pulse_user_text(),
                    }
                )
                await self._bus.publish(
                    {
                        "topic": "task",
                        "task_id": task_id,
                        "event": "log",
                        "line": "[子 Agent 定时心跳] 仍有未完成子任务，继续协调",
                    }
                )
                continue
        except Exception as e:
            if slog:
                slog.write_exception(
                    "task",
                    "AgentLoop",
                    "sub_agent_worker",
                    e,
                    f"task_id={task_id}",
                )
            raise
        finally:
            current_task_id.reset(token)

    async def chat(self, user_text: str) -> str:
        slog = self._session_log
        if slog:
            u = user_text if len(user_text) <= 12000 else user_text[:11997] + "..."
            slog.write_event("agent", "AgentLoop", "chat_user", u, "")
        await self._store.append_chat_message(
            self._conn, session_id=self._session_id, role="user", content=user_text
        )
        self._messages.append({"role": "user", "content": user_text})
        await self._prepend_main_inbox_if_any()
        tools = openai_tools()
        try:
            for _ in range(self._cfg.agent.max_tool_rounds):
                await self._prepend_main_inbox_if_any()
                batch = [{"role": "system", "content": self._system_content()}, *self._messages]
                await self._bus.publish(
                    {
                        "topic": "agent",
                        "event": "phase",
                        "phase": "llm",
                        "detail": "等待大模型响应…",
                    }
                )
                resp = await self._llm.chat(batch, tools)
                self._log_llm_usage(resp)
                msg = resp.choices[0].message
                if not msg.tool_calls:
                    text = msg.content or ""
                    self._messages.append({"role": "assistant", "content": text})
                    await self._store.append_chat_message(
                        self._conn,
                        session_id=self._session_id,
                        role="assistant",
                        content=text,
                    )
                    if slog:
                        t2 = text if len(text) <= 12000 else text[:11997] + "..."
                        slog.write_event("agent", "AgentLoop", "chat_assistant", "", t2)
                    return text
                ad = _msg_assistant_dict(msg)
                self._messages.append(ad)
                await self._store.append_chat_message(
                    self._conn,
                    session_id=self._session_id,
                    role="assistant",
                    content=msg.content or "",
                    tool_calls=tool_calls_to_jsonable(msg.tool_calls),
                )
                for tc in msg.tool_calls:
                    name = tc.function.name
                    raw_args = tc.function.arguments or "{}"
                    await self._bus.publish(
                        {
                            "topic": "agent",
                            "event": "phase",
                            "phase": "tool",
                            "name": name,
                            "detail": f"执行工具 {name}…",
                        }
                    )
                    result = await self._router.run(name, raw_args)
                    await self._bus.publish(
                        {
                            "topic": "agent",
                            "event": "tool",
                            "name": name,
                            "preview": str(result)[:500],
                        }
                    )
                    body = json.dumps(result, ensure_ascii=False)[:24_000]
                    self._messages.append(
                        {
                            "role": "tool",
                            "tool_call_id": tc.id,
                            "content": body,
                        }
                    )
                    await self._store.append_chat_message(
                        self._conn,
                        session_id=self._session_id,
                        role="tool",
                        content=body,
                        tool_call_id=tc.id,
                        tool_name=name,
                    )
                if self._strip_chat_prefix_after_tool_round:
                    _trim_chat_messages_to_last_user_turn(self._messages)
                    self._strip_chat_prefix_after_tool_round = False
            if slog:
                slog.write_event(
                    "agent",
                    "AgentLoop",
                    "chat_stopped",
                    "",
                    "too_many_tool_rounds",
                )
            return "Stopped: too many tool rounds."
        except Exception as e:
            if slog:
                slog.write_exception("agent", "AgentLoop", "chat", e)
            raise


def merge_env_api_key(cfg: AppConfig) -> None:
    if cfg.llm.api_key:
        return
    proto = (cfg.llm.proto or "openai").strip().lower()
    if proto == "dashscope":
        cfg.llm.api_key = os.environ.get("DASHSCOPE_API_KEY", "") or os.environ.get(
            "OPENAI_API_KEY", ""
        )
    elif proto == "ollama":
        cfg.llm.api_key = (
            os.environ.get("OLLAMA_API_KEY", "")
            or os.environ.get("OPENAI_API_KEY", "")
            or "ollama"
        )
    else:
        cfg.llm.api_key = os.environ.get("OPENAI_API_KEY", "")



from __future__ import annotations

from typing import Any

from aicd.env_probe import (
    build_environment_digest,
    capabilities_summary_path,
    load_runtime_env_snapshot,
    refresh_runtime_env_file,
    runtime_env_path,
    summarize_commands,
)


def _runtime_env_refresh(router: Any, args: dict[str, Any]) -> dict[str, Any]:
    del args  # 无参数；保留以兼容 OpenAI tools schema
    path, data = refresh_runtime_env_file(router._layout)
    router.notify_env_capabilities_updated(data, path)
    avail, total = summarize_commands(data)
    return {
        "ok": True,
        "path": str(path),
        "capabilities_summary_path": str(capabilities_summary_path(router._layout)),
        "generated_at": data.get("generated_at"),
        "commands_available": avail,
        "commands_total": total,
        "hint": "Use fs_read_file on path for full YAML; capabilities_summary_path mirrors the model-facing "
        "snapshot; call again after user installs tools.",
    }


def _agent_environment_digest(router: Any, args: dict[str, Any]) -> dict[str, Any]:
    layout = router._layout
    refresh_first = bool(args.get("refresh_first", False))
    if refresh_first:
        path, data = refresh_runtime_env_file(layout)
        router.notify_env_capabilities_updated(data, path)
    else:
        data = load_runtime_env_snapshot(layout)

    cfg = router._cfg
    yp = str(runtime_env_path(layout).resolve())
    cap_p = str(capabilities_summary_path(layout).resolve())
    ai_out, ai_tmp = router.get_ai_paths()
    digest = build_environment_digest(
        data,
        ai_workspace_cwd=str(router.cwd.resolve()),
        ai_output_root=str(ai_out),
        ai_tmp_dir=str(ai_tmp),
        runtime_yaml_path=yp,
        capabilities_summary_path_str=cap_p,
        agent_max_tool_rounds=int(cfg.agent.max_tool_rounds),
        sub_agent_max_tool_rounds=int(cfg.agent.sub_agent_max_tool_rounds),
        llm_proto=str(cfg.llm.proto),
        llm_model=str(cfg.llm.model),
        llm_api_key_configured=bool(str(cfg.llm.api_key or "").strip()),
    )
    avail, total = summarize_commands(data)
    out: dict[str, Any] = {
        "ok": True,
        "generated_at": data.get("generated_at"),
        "commands_available": avail,
        "commands_total": total,
        **digest,
    }
    if args.get("include_capabilities_text"):
        cap = capabilities_summary_path(layout)
        mc = int(args.get("capabilities_max_chars", 6000) or 6000)
        mc = max(500, min(12000, mc))
        if cap.is_file():
            txt = cap.read_text(encoding="utf-8", errors="replace")
            out["capabilities_summary_excerpt"] = txt[:mc]
        else:
            out["capabilities_summary_excerpt"] = ""
    return out


HANDLERS: dict[str, Any] = {
    "runtime_env_refresh": _runtime_env_refresh,
    "agent_environment_digest": _agent_environment_digest,
}



"""JSON / HTML / XML 处理与轻量结构转换、文本搜索。"""

from __future__ import annotations

import html
import json
import re
import xml.etree.ElementTree as ET
from html.parser import HTMLParser
from typing import Any

MAX_TEXT = 1_000_000
MAX_JSON_DEPTH = 80
MAX_XML_NODES = 5000


class _StripHTMLParser(HTMLParser):
    def __init__(self) -> None:
        super().__init__()
        self._chunks: list[str] = []

    def handle_data(self, data: str) -> None:
        self._chunks.append(data)

    def get_text(self) -> str:
        return " ".join(" ".join(self._chunks).split())


def op_json_parse(payload: dict[str, Any]) -> dict[str, Any]:
    s = str(payload.get("text") or "")
    if len(s) > MAX_TEXT:
        return {"ok": False, "error": f"text exceeds {MAX_TEXT} chars"}
    try:
        obj = json.loads(s)
    except json.JSONDecodeError as e:
        return {"ok": False, "error": str(e)}
    return {"ok": True, "value": obj}


def op_json_stringify(payload: dict[str, Any]) -> dict[str, Any]:
    obj = payload.get("value")
    if obj is None:
        jt = payload.get("json_text")
        if jt is not None:
            try:
                obj = json.loads(str(jt))
            except json.JSONDecodeError as e:
                return {"ok": False, "error": f"json_text: {e}"}
        else:
            return {
                "ok": False,
                "error": "pass value (object) or json_text (JSON string to re-format)",
            }
    try:
        indent = payload.get("indent")
        ind = int(indent) if indent is not None else None
    except (TypeError, ValueError):
        ind = None
    if ind is not None:
        ind = max(0, min(8, ind))
    try:
        s = json.dumps(obj, ensure_ascii=False, indent=ind, default=str)
    except (TypeError, ValueError) as e:
        return {"ok": False, "error": str(e)}
    if len(s) > MAX_TEXT:
        return {"ok": False, "error": "JSON output too large for tool cap"}
    return {"ok": True, "text": s}


def op_json_pretty(payload: dict[str, Any]) -> dict[str, Any]:
    s = str(payload.get("text") or "")
    if len(s) > MAX_TEXT:
        return {"ok": False, "error": f"text exceeds {MAX_TEXT} chars"}
    try:
        obj = json.loads(s)
    except json.JSONDecodeError as e:
        return {"ok": False, "error": str(e)}
    try:
        indent = int(payload.get("indent", 2))
    except (TypeError, ValueError):
        indent = 2
    indent = max(0, min(8, indent))
    out = json.dumps(obj, ensure_ascii=False, indent=indent, default=str)
    if len(out) > MAX_TEXT:
        return {"ok": False, "error": "pretty output too large"}
    return {"ok": True, "text": out}


def _get_by_path(obj: Any, path: str) -> tuple[Any | None, str | None]:
    parts = [p for p in path.split(".") if p]
    cur: Any = obj
    for p in parts:
        if isinstance(cur, dict):
            if p not in cur:
                return None, f"missing key {p!r}"
            cur = cur[p]
        elif isinstance(cur, list):
            try:
                idx = int(p)
            except ValueError:
                return None, f"expected int index in list, got {p!r}"
            if idx < 0 or idx >= len(cur):
                return None, f"index {idx} out of range (len={len(cur)})"
            cur = cur[idx]
        else:
            return None, f"cannot descend into {type(cur).__name__} at {p!r}"
    return cur, None


def op_json_path_get(payload: dict[str, Any]) -> dict[str, Any]:
    s = str(payload.get("text") or "")
    path = str(payload.get("path") or "")
    if not path:
        return {"ok": False, "error": "path required (dot-separated, e.g. a.b.0)"}
    if len(s) > MAX_TEXT:
        return {"ok": False, "error": "text too large"}
    try:
        obj = json.loads(s)
    except json.JSONDecodeError as e:
        return {"ok": False, "error": str(e)}
    val, err = _get_by_path(obj, path)
    if err:
        return {"ok": False, "error": err}
    if isinstance(val, (dict, list)):
        snippet = json.dumps(val, ensure_ascii=False, default=str)
        if len(snippet) > 120_000:
            return {
                "ok": True,
                "type": type(val).__name__,
                "truncated": True,
                "preview": snippet[:50_000] + "…",
            }
        return {"ok": True, "type": type(val).__name__, "value": val}
    return {"ok": True, "type": type(val).__name__, "value": val}


def op_html_to_text(payload: dict[str, Any]) -> dict[str, Any]:
    s = str(payload.get("html") or payload.get("text") or "")
    if len(s) > MAX_TEXT:
        return {"ok": False, "error": f"html exceeds {MAX_TEXT} chars"}
    parser = _StripHTMLParser()
    try:
        parser.feed(s)
        parser.close()
    except Exception as e:  # noqa: BLE001
        return {"ok": False, "error": f"parse: {e}"}
    text = parser.get_text()
    if payload.get("unescape_entities"):
        text = html.unescape(text)
    return {"ok": True, "text": text}


def op_html_escape(payload: dict[str, Any]) -> dict[str, Any]:
    s = str(payload.get("text") or "")
    if len(s) > MAX_TEXT:
        return {"ok": False, "error": "text too large"}
    return {"ok": True, "text": html.escape(s, quote=True)}


def op_xml_parse_flat(payload: dict[str, Any]) -> dict[str, Any]:
    s = str(payload.get("xml") or payload.get("text") or "")
    if len(s) > MAX_TEXT:
        return {"ok": False, "error": "xml too large"}
    try:
        root = ET.fromstring(s)
    except ET.ParseError as e:
        return {"ok": False, "error": str(e)}

    rows: list[dict[str, Any]] = []

    def walk(el: ET.Element, depth: int) -> None:
        if len(rows) >= MAX_XML_NODES:
            return
        if depth > MAX_JSON_DEPTH:
            return
        tag = el.tag.split("}")[-1] if "}" in el.tag else el.tag
        text = (el.text or "").strip()
        attrs = {k.split("}")[-1]: v for k, v in el.attrib.items()}
        children = list(el)
        rows.append(
            {
                "tag": tag,
                "depth": depth,
                "text": text[:2000] if text else "",
                "attrib": attrs,
                "child_tags": [c.tag.split("}")[-1] for c in children[:20]],
            }
        )
        for c in children:
            walk(c, depth + 1)

    walk(root, 0)
    return {
        "ok": True,
        "root_tag": root.tag.split("}")[-1],
        "nodes": rows,
        "truncated": len(rows) >= MAX_XML_NODES,
        "note": "trusted XML only; no XXE hardening",
    }


def op_xml_find_tags(payload: dict[str, Any]) -> dict[str, Any]:
    s = str(payload.get("xml") or payload.get("text") or "")
    name = str(payload.get("tag") or "")
    if not name:
        return {"ok": False, "error": "tag required"}
    if len(s) > MAX_TEXT:
        return {"ok": False, "error": "xml too large"}
    try:
        root = ET.fromstring(s)
    except ET.ParseError as e:
        return {"ok": False, "error": str(e)}
    found: list[dict[str, Any]] = []
    for el in root.iter():
        t = el.tag.split("}")[-1]
        if t == name or name == "*":
            txt = "".join(el.itertext()).strip()
            found.append(
                {
                    "tag": t,
                    "text": txt[:5000],
                    "attrib": {k.split("}")[-1]: v for k, v in el.attrib.items()},
                }
            )
            if len(found) >= 500:
                break
    return {"ok": True, "matches": found, "count": len(found)}


def op_structure_convert(payload: dict[str, Any]) -> dict[str, Any]:
    """list[dict] -> dict keyed by field; sort_keys on dict."""
    op = str(payload.get("mode") or "").strip().lower()
    obj = payload.get("value")
    if op == "list_to_dict":
        if not isinstance(obj, list):
            return {"ok": False, "error": "value must be JSON array"}
        key_field = str(payload.get("key_field") or "id")
        out: dict[str, Any] = {}
        dup = 0
        for i, item in enumerate(obj):
            if not isinstance(item, dict):
                return {"ok": False, "error": f"item {i} is not object"}
            k = item.get(key_field)
            if k is None:
                return {"ok": False, "error": f"missing {key_field} at index {i}"}
            ks = str(k)
            if ks in out:
                dup += 1
            out[ks] = item
        return {"ok": True, "dict": out, "duplicate_keys_overwritten": dup}
    if op == "sort_object_keys":
        if not isinstance(obj, dict):
            return {"ok": False, "error": "value must be JSON object"}

        def sort_obj(o: Any) -> Any:
            if isinstance(o, dict):
                return {k: sort_obj(o[k]) for k in sorted(o.keys(), key=str)}
            if isinstance(o, list):
                return [sort_obj(x) for x in o]
            return o

        return {"ok": True, "value": sort_obj(obj)}
    if op == "transpose_records":
        if not isinstance(obj, list) or not obj:
            return {"ok": False, "error": "value must be non-empty array of objects"}
        keys: set[str] = set()
        for it in obj:
            if isinstance(it, dict):
                keys.update(it.keys())
        cols = sorted(keys)
        series: dict[str, list[Any]] = {c: [] for c in cols}
        for it in obj:
            if not isinstance(it, dict):
                return {"ok": False, "error": "all items must be objects"}
            for c in cols:
                series[c].append(it.get(c))
        return {"ok": True, "columns": cols, "series": series, "rows": len(obj)}
    return {
        "ok": False,
        "error": f"unknown mode: {op!r}",
        "hint": "list_to_dict|sort_object_keys|transpose_records",
    }


def op_text_find(payload: dict[str, Any]) -> dict[str, Any]:
    text = str(payload.get("text") or "")
    if len(text) > MAX_TEXT:
        return {"ok": False, "error": "text too large"}
    sub = str(payload.get("substring") or "")
    if sub:
        positions = []
        start = 0
        max_hits = int(payload.get("max_hits", 200))
        max_hits = max(1, min(2000, max_hits))
        while len(positions) < max_hits:
            i = text.find(sub, start)
            if i < 0:
                break
            positions.append(i)
            start = i + max(1, len(sub))
        return {"ok": True, "count": len(positions), "positions": positions}
    pattern = str(payload.get("pattern") or "")
    if not pattern:
        return {"ok": False, "error": "substring or pattern required"}
    flags = 0
    if payload.get("ignore_case"):
        flags |= re.IGNORECASE
    if payload.get("multiline"):
        flags |= re.MULTILINE
    try:
        rx = re.compile(pattern, flags)
    except re.error as e:
        return {"ok": False, "error": str(e)}
    max_hits = int(payload.get("max_hits", 100))
    max_hits = max(1, min(2000, max_hits))
    matches: list[dict[str, Any]] = []
    for i, m in enumerate(rx.finditer(text)):
        if i >= max_hits:
            break
        matches.append(
            {
                "start": m.start(),
                "end": m.end(),
                "match": m.group(0)[:500],
            }
        )
    return {"ok": True, "matches": matches, "count": len(matches)}


def run_operation(operation: str, payload: dict[str, Any]) -> dict[str, Any]:
    op = operation.strip().lower().replace("-", "_")
    if op in ("json_parse",):
        return op_json_parse(payload)
    if op in ("json_stringify", "json_dumps"):
        return op_json_stringify(payload)
    if op in ("json_pretty", "json_format"):
        return op_json_pretty(payload)
    if op in ("json_path_get", "json_get"):
        return op_json_path_get(payload)
    if op in ("html_to_text", "html_strip"):
        return op_html_to_text(payload)
    if op in ("html_escape",):
        return op_html_escape(payload)
    if op in ("xml_parse_flat", "xml_flat"):
        return op_xml_parse_flat(payload)
    if op in ("xml_find_tags", "xml_find"):
        return op_xml_find_tags(payload)
    if op in ("structure_convert", "struct_convert"):
        return op_structure_convert(payload)
    if op in ("text_find", "find_text"):
        return op_text_find(payload)
    return {
        "ok": False,
        "error": f"unknown operation: {operation!r}",
        "hint": "json_parse|json_stringify|json_pretty|json_path_get|html_to_text|html_escape|"
        "xml_parse_flat|xml_find_tags|structure_convert|text_find",
    }



"""AgentLoop 新会话后消息链与裁剪逻辑。"""

from __future__ import annotations

from aicd.agent.loop import _trim_chat_messages_to_last_user_turn


def test_trim_keeps_suffix_from_last_user() -> None:
    msgs = [
        {"role": "user", "content": "old"},
        {"role": "assistant", "content": "a"},
        {"role": "user", "content": "新会话"},
        {"role": "assistant", "content": "", "tool_calls": []},
        {"role": "tool", "tool_call_id": "x", "content": "{}"},
    ]
    _trim_chat_messages_to_last_user_turn(msgs)
    assert len(msgs) == 3
    assert msgs[0]["content"] == "新会话"


def test_trim_noop_when_only_one_user_at_start() -> None:
    msgs = [
        {"role": "user", "content": "hi"},
        {"role": "assistant", "content": "", "tool_calls": []},
    ]
    _trim_chat_messages_to_last_user_turn(msgs)
    assert len(msgs) == 2



from __future__ import annotations

import json
from datetime import UTC, datetime
from typing import Any


def now_iso(*, use_utc: bool = True, tz_name: str | None = None) -> dict[str, Any]:
    if tz_name:
        try:
            from zoneinfo import ZoneInfo

            tz = ZoneInfo(tz_name)
        except Exception as e:
            return {"ok": False, "error": f"invalid timezone: {e}"}
        dt = datetime.now(tz)
        return {"ok": True, "iso": dt.isoformat(), "timezone": tz_name}
    if use_utc:
        dt = datetime.now(UTC)
        return {"ok": True, "iso": dt.isoformat(), "timezone": "UTC"}
    dt = datetime.now().astimezone()
    return {"ok": True, "iso": dt.isoformat(), "timezone": str(dt.tzinfo)}


def _parse_iso(s: str) -> datetime:
    s = s.strip()
    if s.endswith("Z"):
        s = s[:-1] + "+00:00"
    return datetime.fromisoformat(s)


def elapsed_ms(start_iso: str, end_iso: str) -> dict[str, Any]:
    try:
        a = _parse_iso(start_iso)
        b = _parse_iso(end_iso)
    except ValueError as e:
        return {"ok": False, "error": f"invalid iso timestamp: {e}"}
    delta = b - a
    ms = int(delta.total_seconds() * 1000)
    return {"ok": True, "duration_ms": ms}


def log_record(
    level: str,
    message: str,
    **fields: Any,
) -> dict[str, Any]:
    rec: dict[str, Any] = {
        "level": level,
        "message": message,
        **fields,
    }
    now = now_iso(use_utc=True)
    if now.get("ok"):
        rec["ts"] = now["iso"]
    return {"ok": True, "record": rec}


def append_jsonl(path: str, record: dict[str, Any]) -> dict[str, Any]:
    line = json.dumps(record, ensure_ascii=False) + "\n"
    with open(path, "a", encoding="utf-8") as f:
        f.write(line)
    return {"ok": True, "path": path, "bytes": len(line.encode("utf-8"))}


def run_operation(operation: str, payload: dict[str, Any]) -> dict[str, Any]:
    op = operation.strip().lower()
    if op == "now_iso":
        return now_iso(
            use_utc=bool(payload.get("use_utc", True)),
            tz_name=payload.get("timezone") or None,
        )
    if op == "elapsed_ms":
        return elapsed_ms(
            str(payload.get("start_iso") or ""),
            str(payload.get("end_iso") or ""),
        )
    if op == "log_record":
        msg = str(payload.get("message") or "")
        lvl = str(payload.get("level") or "info")
        extra = {k: v for k, v in payload.items() if k not in ("message", "level", "operation")}
        return log_record(lvl, msg, **extra)
    return {"ok": False, "error": f"unknown operation: {operation}"}



"""启动环境横幅文案。"""

from __future__ import annotations

from aicd.env_probe import (
    BANNER_CRITICAL,
    BANNER_RECOMMEND,
    format_startup_env_chat_banner,
    startup_env_chat_banner_lines,
)


def test_banner_mentions_llm_when_no_key() -> None:
    data = {
        "os": {"system": "Windows", "release": "11"},
        "python": {"version": "3.12.0"},
        "commands": {"git": {"available": True}, "node": {"available": False}},
        "python_modules_optional": {},
    }
    s = format_startup_env_chat_banner(data, has_llm_api_key=False)
    assert "LLM" in s or "密钥" in s
    assert "系统环境检查" in s


def test_banner_ok_with_key() -> None:
    data = {
        "os": {"system": "Linux"},
        "python": {"version": "3.11"},
        "commands": {f"k{i}": {"available": True} for i in range(5)},
        "python_modules_optional": {"pytesseract": {"available": False}},
    }
    s = format_startup_env_chat_banner(data, has_llm_api_key=True)
    assert "runtime_env_refresh" in s
    assert "pytesseract" in s


def test_banner_lines_critical_without_llm_key() -> None:
    data = {
        "os": {"system": "Windows", "release": "11"},
        "python": {"version": "3.12.0"},
        "commands": {"git": {"available": True}},
        "python_modules_optional": {},
    }
    rows = startup_env_chat_banner_lines(data, has_llm_api_key=False)
    kinds = [k for _, k in rows]
    assert BANNER_CRITICAL in kinds
    assert BANNER_RECOMMEND in kinds


def test_banner_lines_no_critical_when_key_ok() -> None:
    data = {
        "os": {"system": "Linux"},
        "python": {"version": "3.11"},
        "commands": {"git": {"available": True}, "node": {"available": True}},
        "python_modules_optional": {},
    }
    rows = startup_env_chat_banner_lines(data, has_llm_api_key=True)
    kinds = [k for _, k in rows]
    assert BANNER_CRITICAL not in kinds



"""Markdown → PDF：默认 **PyMuPDF Story**（MD→HTML→排版）；可选 **pandoc** 外部引擎。"""

from __future__ import annotations

import shutil
import subprocess
from pathlib import Path
from typing import Any, Callable

PolicyCheck = Callable[[Path], None]

# PyMuPDF Story 的 HTML/CSS 子集较弱：thead/th 上隐式或继承的 background 常被画成**整块黑色**；
# border-collapse:collapse 与粗线也可能变成异常横条。以下用**显式浅色**、separate 边框、细 hr。
DEFAULT_STORY_CSS = """
body { font-family: Helvetica, Arial, "Microsoft YaHei", "PingFang SC", sans-serif; font-size: 11pt; line-height: 1.35; color: #000000; background-color: #ffffff; }
h1 { font-size: 20pt; margin-top: 0.6em; margin-bottom: 0.3em; color: #000000; background-color: transparent; }
h2 { font-size: 16pt; margin-top: 0.6em; margin-bottom: 0.3em; color: #000000; background-color: transparent; }
h3 { font-size: 13pt; margin-top: 0.5em; margin-bottom: 0.25em; color: #000000; background-color: transparent; }
pre { background-color: #f4f4f4; color: #000000; padding: 8px; font-size: 9pt; overflow-x: auto; border: 0.5pt solid #dddddd; }
code { font-family: Consolas, monospace; font-size: 9.5pt; background-color: #f4f4f4; color: #000000; padding: 1px 4px; }
table { border-collapse: separate; border-spacing: 0; width: 100%; margin: 0.5em 0; background-color: #ffffff; }
thead { background-color: #e8e8e8; color: #000000; }
tbody { background-color: #ffffff; }
tr { background-color: #ffffff; }
thead tr { background-color: #e8e8e8; }
th, td {
  border: 0.75pt solid #888888;
  padding: 4px 6px;
  vertical-align: top;
  background-color: #ffffff;
  color: #000000;
}
th {
  background-color: #eeeeee;
  font-weight: bold;
  color: #000000;
}
thead th { background-color: #e8e8e8; color: #000000; }
blockquote { border-left: 3px solid #cccccc; margin-left: 0; padding-left: 12px; color: #444444; background-color: transparent; }
img { max-width: 100%; height: auto; }
hr {
  display: block;
  height: 0;
  border: 0;
  border-top: 0.75pt solid #999999;
  margin: 0.8em 0;
  padding: 0;
  background-color: transparent;
  color: #000000;
}
"""

MAX_MARKDOWN_CHARS = 2_000_000


def _read_markdown(
    markdown_path: Path | None, markdown_text: str | None
) -> tuple[str | None, str | None]:
    if markdown_path is not None and markdown_text is not None:
        return None, "pass only one of markdown_path or markdown_text"
    if markdown_path is None and (markdown_text is None or not str(markdown_text).strip()):
        return None, "markdown_path or markdown_text is required"
    if markdown_path is not None:
        p = markdown_path
        if not p.is_file():
            return None, f"markdown file not found: {p}"
        if p.suffix.lower() not in (".md", ".markdown", ".txt"):
            return None, "markdown_path should be .md / .markdown / .txt"
        raw = p.read_text(encoding="utf-8", errors="replace")
        if len(raw) > MAX_MARKDOWN_CHARS:
            return None, f"markdown exceeds {MAX_MARKDOWN_CHARS} characters"
        return raw, None
    t = str(markdown_text or "")
    if len(t) > MAX_MARKDOWN_CHARS:
        return None, f"markdown exceeds {MAX_MARKDOWN_CHARS} characters"
    return t, None


def _md_to_html_fragment(md: str) -> tuple[str | None, str | None]:
    """延迟导入 ``markdown``，避免未 ``pip install`` 时阻塞整个 Agent（ToolRouter 加载）。"""
    try:
        import markdown as md_mod
    except ImportError:
        return (
            None,
            "缺少 Python 包 markdown（ModuleNotFoundError）。请执行: pip install \"markdown>=3.5\" "
            "或重新 pip install -e . 后重启 Aicd。",
        )
    exts = [
        "markdown.extensions.tables",
        "markdown.extensions.fenced_code",
        "markdown.extensions.nl2br",
        "markdown.extensions.sane_lists",
    ]
    return md_mod.markdown(md, extensions=exts), None


def export_markdown_to_pdf_pymupdf(
    md_source: str,
    output_pdf: Path,
    *,
    policy_check: PolicyCheck,
    archive_dir: Path,
    user_css: str | None = None,
    paper: str = "a4",
    margin_pt: float = 36.0,
) -> dict[str, Any]:
    import fitz

    output_pdf = output_pdf.expanduser().resolve()
    policy_check(output_pdf)
    policy_check(output_pdf.parent)
    ad = archive_dir.expanduser().resolve()
    policy_check(ad)

    fragment, md_err = _md_to_html_fragment(md_source)
    if md_err or fragment is None:
        return {
            "ok": False,
            "error": md_err or "markdown to HTML failed",
            "engine": "pymupdf_story",
            "hint": "pip install \"markdown>=3.5\"",
        }
    html = (
        "<!DOCTYPE html><html><head><meta charset='utf-8'/></head><body>"
        + fragment
        + "</body></html>"
    )
    css = (user_css or "").strip() + "\n" + DEFAULT_STORY_CSS
    story = fitz.Story(html=html, user_css=css, archive=str(ad))
    mediabox = fitz.paper_rect(paper.lower())
    m = max(12.0, float(margin_pt))
    where = mediabox + (m, m, -m, -m)

    output_pdf.parent.mkdir(parents=True, exist_ok=True)
    writer = fitz.DocumentWriter(str(output_pdf))
    try:
        more = 1
        while more:
            dev = writer.begin_page(mediabox)
            more, _ = story.place(where)
            story.draw(dev)
            writer.end_page()
        writer.close()
    except Exception as e:  # noqa: BLE001
        try:
            writer.close()
        except Exception:  # noqa: BLE001
            pass
        if output_pdf.is_file():
            try:
                output_pdf.unlink()
            except OSError:
                pass
        return {"ok": False, "error": repr(e), "engine": "pymupdf_story"}

    if not output_pdf.is_file() or output_pdf.stat().st_size == 0:
        return {"ok": False, "error": "empty PDF output", "engine": "pymupdf_story"}
    return {
        "ok": True,
        "pdf_path": str(output_pdf),
        "engine": "pymupdf_story",
        "paper": paper.lower(),
    }


def export_markdown_to_pdf_pandoc(
    md_path: Path,
    output_pdf: Path,
    *,
    policy_check: PolicyCheck,
    extra_args: list[str] | None = None,
) -> dict[str, Any]:
    exe = shutil.which("pandoc")
    if not exe:
        return {
            "ok": False,
            "error": "pandoc not found on PATH",
            "hint": "Install Pandoc (https://pandoc.org) or use engine=pymupdf (built-in).",
            "engine": "pandoc",
        }
    src = md_path.expanduser().resolve()
    dst = output_pdf.expanduser().resolve()
    policy_check(src)
    policy_check(dst)
    policy_check(dst.parent)
    if not src.is_file():
        return {"ok": False, "error": f"source not found: {src}", "engine": "pandoc"}
    dst.parent.mkdir(parents=True, exist_ok=True)
    args = [exe, str(src), "-o", str(dst)]
    if extra_args:
        args.extend(str(x) for x in extra_args if str(x).strip())
    try:
        cp = subprocess.run(
            args,
            capture_output=True,
            text=True,
            timeout=600,
            encoding="utf-8",
            errors="replace",
            cwd=str(src.parent),
        )
    except subprocess.TimeoutExpired:
        return {"ok": False, "error": "pandoc timed out (600s)", "engine": "pandoc"}
    except OSError as e:
        return {"ok": False, "error": f"pandoc failed to run: {e}", "engine": "pandoc"}
    if cp.returncode != 0:
        tail = (cp.stderr or cp.stdout or "")[:4000]
        return {
            "ok": False,
            "error": f"pandoc exit {cp.returncode}",
            "stderr_tail": tail,
            "engine": "pandoc",
        }
    if not dst.is_file() or dst.stat().st_size == 0:
        return {"ok": False, "error": "pandoc produced no PDF", "engine": "pandoc"}
    return {"ok": True, "pdf_path": str(dst), "engine": "pandoc"}


def _write_source_snapshot(
    *,
    snapshot_path: Path,
    policy_check: PolicyCheck,
    markdown_path: Path | None,
    raw_text: str | None,
) -> str | None:
    """将参与排版的 Markdown 落盘，便于日后只改编排不改内容。返回绝对路径或 None。"""
    sp = snapshot_path.expanduser().resolve()
    policy_check(sp)
    policy_check(sp.parent)
    sp.parent.mkdir(parents=True, exist_ok=True)
    if markdown_path is not None and Path(markdown_path).is_file():
        shutil.copy2(Path(markdown_path).resolve(), sp)
    elif raw_text is not None:
        sp.write_text(raw_text, encoding="utf-8")
    else:
        return None
    return str(sp)


def export_markdown_to_pdf(
    *,
    output_pdf_path: Path,
    policy_check: PolicyCheck,
    markdown_path: Path | None = None,
    markdown_text: str | None = None,
    engine: str = "pymupdf",
    archive_path: Path | None = None,
    user_css: str | None = None,
    paper: str = "a4",
    margin_pt: float = 36.0,
    pandoc_extra_args: list[str] | None = None,
    source_snapshot_path: Path | None = None,
) -> dict[str, Any]:
    """
    ``engine``: ``pymupdf`` | ``pandoc``.

    - **pymupdf**：内置 **Story**；``archive_path`` 为解析 ``![](relative.png)`` 的基准目录（默认：``markdown_path`` 父目录或 workspace 当前策略路径）。
    - **pandoc**：需系统安装 **pandoc**；仅支持 ``markdown_path``（临时文件可在外部写入）。
    - **source_snapshot_path**：可选；将本次用于生成的 **.md** 副本写到该路径（与 PDF 同目录的 ``*_source.md`` 便于改版式不改内容）。
    """
    dst = output_pdf_path.expanduser().resolve()
    if dst.suffix.lower() != ".pdf":
        return {"ok": False, "error": "output_pdf_path must end with .pdf"}

    eng = (engine or "pymupdf").strip().lower()
    raw, err = _read_markdown(
        Path(markdown_path) if markdown_path is not None else None,
        markdown_text,
    )
    if err:
        return {"ok": False, "error": err}

    if eng == "pandoc":
        if markdown_path is None:
            return {
                "ok": False,
                "error": "engine=pandoc requires markdown_path on disk",
                "hint": "Save markdown to a workspace file first, or use engine=pymupdf with markdown_text.",
            }
        out = export_markdown_to_pdf_pandoc(
            Path(markdown_path),
            dst,
            policy_check=policy_check,
            extra_args=pandoc_extra_args,
        )
        if out.get("ok") and source_snapshot_path is not None:
            snap = _write_source_snapshot(
                snapshot_path=source_snapshot_path,
                policy_check=policy_check,
                markdown_path=Path(markdown_path),
                raw_text=None,
            )
            if snap:
                out["source_snapshot_path"] = snap
        return out

    if eng not in ("pymupdf", "pymupdf_story", ""):
        return {
            "ok": False,
            "error": f"unknown engine: {engine!r}; use pymupdf or pandoc",
        }

    base = archive_path
    if base is None:
        base = Path(markdown_path).resolve().parent if markdown_path else dst.parent
    base = Path(base).expanduser().resolve()
    policy_check(base)

    out = export_markdown_to_pdf_pymupdf(
        raw or "",
        dst,
        policy_check=policy_check,
        archive_dir=base,
        user_css=user_css,
        paper=paper,
        margin_pt=margin_pt,
    )
    if out.get("ok") and source_snapshot_path is not None:
        snap = _write_source_snapshot(
            snapshot_path=source_snapshot_path,
            policy_check=policy_check,
            markdown_path=Path(markdown_path) if markdown_path else None,
            raw_text=raw,
        )
        if snap:
            out["source_snapshot_path"] = snap
    return out


def pandoc_available() -> bool:
    return bool(shutil.which("pandoc"))



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from aicd.config import save_config
from aicd.config_settings import apply_nested_settings_patch, format_settings_summary

if TYPE_CHECKING:
    from aicd.agent.tool_router import ToolRouter


def _agent_settings_update(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    patch = args.get("patch")
    if patch is None:
        return {
            "ok": True,
            "message": "no patch; current summary",
            "summary": format_settings_summary(router._cfg),
        }
    if not isinstance(patch, dict):
        return {"ok": False, "error": "patch must be an object with optional llm / agent keys"}
    ok, msg = apply_nested_settings_patch(router._cfg, patch)
    if not ok:
        return {"ok": False, "error": msg}
    if bool(args.get("persist", True)):
        save_config(router._layout["config"], router._cfg)
    return {
        "ok": True,
        "message": msg,
        "summary": format_settings_summary(router._cfg),
        "persisted": bool(args.get("persist", True)),
    }


HANDLERS: dict[str, Any] = {
    "agent_settings_update": _agent_settings_update,
}



"""主对话与子 Agent 的「无用户输入」定时唤醒文案（TUI 心跳 / 子任务协调）。"""

from __future__ import annotations

# 与 TUI 侧识别一致：勿改前缀，否则 _chat_pump_loop 无法标记系统行
MAIN_IDLE_PULSE_PREFIX = "（Aicd·主对话·定时心跳）"


def is_main_idle_pulse(user_text: str) -> bool:
    return user_text.lstrip().startswith(MAIN_IDLE_PULSE_PREFIX)


def main_idle_pulse_user_text() -> str:
    return (
        MAIN_IDLE_PULSE_PREFIX
        + " 当前无新的真人输入。请先 **task_main_inbox**（或依赖运行时已注入的【主收件箱】）与 **task_session_children_digest** "
        "查看本会话顶层子任务是否有 **lifecycle** 完成/失败/取消或其它上行消息；再用 **task_list**、**task_result**、**task_thread_stats** "
        "核对 **running** / **failed** / **pending** 与 **timeline** / **sub_agent_trace**。"
        "若有失败、长期停滞或需要用户决策，用中文在聊天区**简短**说明并视情况 **task_cancel** 或对仍运行的子任务 **task_message_send**（**directive** / **question**）。"
        "检查周期由 **`agent.idleHeartbeatSec`**（TUI **`/agent heartbeat <秒>`**）或 **`agent_settings_update`** 自主设置。"
        "若一切正常、无需打扰用户，可只回复：**·**（中间点）。"
    )


def sub_agent_idle_pulse_user_text() -> str:
    return (
        "（Aicd·子 Agent·定时心跳）收件箱暂无新 **task_message**，但你仍有 **running** 或 **pending** 的子任务（**task_list** 里 **parent_id** 等于你当前 **task_id** 的行）。"
        "请用 **task_result** 查看子任务；需要上级同步时，先 **task_result**（**task_id**=你自己）读 **parent_id**：若非空则 **task_message_send** 的 **to_task_id**=该 **parent_id**，**kind** 用 **status** / **question** / **alert**；"
        "若为顶层子任务（**parent_id** 为空），主对话会通过定时心跳与用户输入查看 **task_list**，你应继续用 **task_progress** 保持可见。"
        "失败子任务可考虑 **task_cancel** 或拆成更小的 **task_ai**。无事可回复 **·**。"
    )



"""image_transform：Pillow 操作与 help；OpenCV 可选。"""

from __future__ import annotations

from pathlib import Path

from PIL import Image

from aicd.config import PathPolicyConfig
from aicd.tools.image_transform import list_image_operations, run_image_transform
from aicd.tools.path_policy import PathPolicy


def test_list_operations_contains_resize() -> None:
    assert "resize" in list_image_operations()
    assert "help" not in list_image_operations()


def test_help_returns_schema(tmp_path: Path) -> None:
    pol = PathPolicy(PathPolicyConfig(full_disk=True))
    r = run_image_transform(
        "help",
        tmp_path / "i.png",
        tmp_path / "o.png",
        params={},
        policy_check=pol.check_allowed,
    )
    assert r["ok"] is True
    assert "operations_help" in r


def test_solid_canvas_no_input_file(tmp_path: Path) -> None:
    pol = PathPolicy(PathPolicyConfig(full_disk=True))
    out = tmp_path / "canvas.png"
    r = run_image_transform(
        "solid_canvas",
        tmp_path / "nonexistent_in.png",
        out,
        params={"width": 120, "height": 80, "fill": [240, 248, 255, 255]},
        policy_check=pol.check_allowed,
    )
    assert r["ok"] is True
    assert out.is_file()
    with Image.open(out) as im:
        assert im.size == (120, 80)


def test_resize_fit(tmp_path: Path) -> None:
    pol = PathPolicy(PathPolicyConfig(full_disk=True))
    src = tmp_path / "a.png"
    Image.new("RGBA", (200, 100), (255, 0, 0, 255)).save(src)
    dst = tmp_path / "b.png"
    r = run_image_transform(
        "resize",
        src,
        dst,
        params={"width": 100, "height": 100, "mode": "fit"},
        policy_check=pol.check_allowed,
    )
    assert r["ok"] is True
    with Image.open(dst) as im:
        assert im.size == (100, 50)


def test_draw_and_overlay(tmp_path: Path) -> None:
    pol = PathPolicy(PathPolicyConfig(full_disk=True))
    base = tmp_path / "base.png"
    Image.new("RGBA", (100, 100), (200, 200, 200, 255)).save(base)
    ov = tmp_path / "ov.png"
    Image.new("RGBA", (20, 20), (0, 0, 255, 128)).save(ov)
    mid = tmp_path / "mid.png"
    r1 = run_image_transform(
        "draw_rectangle",
        base,
        mid,
        params={
            "xyxy": [10, 10, 90, 90],
            "outline": [0, 0, 0, 255],
            "width": 3,
        },
        policy_check=pol.check_allowed,
    )
    assert r1["ok"] is True
    out = tmp_path / "final.png"
    r2 = run_image_transform(
        "paste_overlay",
        mid,
        out,
        params={"overlay_path": str(ov), "x": 40, "y": 40, "alpha": 0.8},
        policy_check=pol.check_allowed,
    )
    assert r2["ok"] is True


def test_edge_canny_if_cv2() -> None:
    try:
        import cv2  # noqa: F401
    except ImportError:
        return
    from tempfile import mkdtemp

    pol = PathPolicy(PathPolicyConfig(full_disk=True))
    d = Path(mkdtemp())
    src = d / "s.png"
    Image.new("RGBA", (64, 64), (128, 128, 128, 255)).save(src)
    dst = d / "e.png"
    r = run_image_transform(
        "edge_canny",
        src,
        dst,
        params={"low": 40, "high": 120},
        policy_check=pol.check_allowed,
    )
    assert r["ok"] is True
    assert dst.stat().st_size > 50



"""TUI 底部输入：多行 TextArea + ``/`` 命令行幽灵补全；单行 Input 保留供兼容。"""

from __future__ import annotations

from textual import events
from textual.binding import Binding
from textual.message import Message
from textual.suggester import Suggester
from textual.widgets import Input, TextArea


# 列表顺序 = 同前缀时 Tab/→ 采纳的优先级（不再按字符串总长最短，避免误优先 /task ai）
_COMPLETIONS: list[str] = [
    "/?",
    "/q",
    "/exit",
    "/quit",
    "/help",
    "/work ",
    "/chat ",
    "/chat stop",
    "/doc ",
    "/llm ",
    "/agent ",
    "/history ",
    "/history list",
    "/history clear",
    "/history copy ",
    "/history help",
    "/task ",
    "/task stop",
    "/task list",
    "/task clear",
    "/task help",
    "/task shell ",
    "/task ai ",
    "/task dummy ",
]

_POS: dict[str, int] = {c: i for i, c in enumerate(_COMPLETIONS)}


def pick_slash_completion(value: str) -> str | None:
    """
    返回比当前输入更长、且以前缀匹配的第一条补全（最短总长优先，其次列表顺序）。
    非 `/` 开头返回 None。
    """
    if not value.startswith("/"):
        return None
    vn = value.casefold()
    matches: list[str] = []
    for c in _COMPLETIONS:
        if len(c) <= len(vn):
            continue
        if c.casefold().startswith(vn):
            matches.append(c)
    if not matches:
        return None
    return min(matches, key=lambda c: _POS[c])


class SlashCommandSuggester(Suggester):
    """仅对以 ``/`` 开头的行给出幽灵补全；普通聊天不提示。"""

    def __init__(self) -> None:
        super().__init__(case_sensitive=False, use_cache=True)

    async def get_suggestion(self, value: str) -> str | None:
        # case_sensitive=False 时 value 已为 casefold
        if not value.startswith("/"):
            return None
        return pick_slash_completion(value)


class ChatInput(Input):
    """单行输入（遗留）：Tab / → 采纳幽灵补全。"""

    BINDINGS = [
        *Input.BINDINGS,
        Binding("tab", "accept_suggestion_tab", "补全", show=False),
    ]

    def action_accept_suggestion_tab(self) -> None:
        if self.cursor_at_end and self._suggestion:
            self.action_cursor_right()


class ChatTextArea(TextArea):
    """多行聊天输入：**Enter** 发送；**Shift+Enter** 换行。

    当前行若以 ``/`` 开头，显示与单行模式相同的幽灵补全，**→** 采纳（与 TextArea 默认一致）。
    """

    BINDINGS = TextArea.BINDINGS

    class Submitted(Message):
        """用户按 Enter 提交整段文本（可含多行）。"""

        def __init__(self, chat: "ChatTextArea", value: str) -> None:
            self.chat = chat
            self.value = value
            super().__init__()

        @property
        def control(self) -> "ChatTextArea":
            return self.chat

    def update_suggestion(self) -> None:
        """仅当光标在某行且该行以 ``/`` 开头时给出补全后缀（与 Input+Suggester 行为对齐）。"""
        row, _col = self.cursor_location
        lines = self.text.split("\n")
        if row < 0 or row >= len(lines):
            self.suggestion = ""
            return
        current_line = lines[row]
        if not current_line.startswith("/"):
            self.suggestion = ""
            return
        comp = pick_slash_completion(current_line)
        if comp and len(comp) > len(current_line):
            self.suggestion = comp[len(current_line) :]
        else:
            self.suggestion = ""

    async def _on_key(self, event: events.Key) -> None:
        """Enter 发送；Shift+Enter 插入换行（与常见 IM 一致）。"""
        if not self.read_only and event.key == "enter":
            if event.shift:
                await super()._on_key(event)
                return
            event.stop()
            event.prevent_default()
            self.action_submit_chat()
            return
        await super()._on_key(event)

    def action_submit_chat(self) -> None:
        raw = self.text
        if not raw.strip():
            return
        self.post_message(self.Submitted(self, raw))
        self.text = ""
        self.suggestion = ""



"""Pure analytics / text / time / chart helpers for Agent tools (JSON-friendly)."""

__all__ = [
    "chart_spec",
    "crypto_codec",
    "drawio_spec",
    "format_data",
    "math_stats",
    "script_syntax",
    "script_templates",
    "text_regex",
    "time_log",
    "validate_viz",
]
